#include <Config.h>
#include <ArrayUtils.h>

#include "../Bass.h"
#include "../db/Database.h"
#include "../itemstructs/ItemSet.h"
#include "../itemstructs/PatternFile.h"

#include "IterativeTop1Collection.h"

IterativeTop1Collection::IterativeTop1Collection(Config *config, Database *db, PatternFile *pFile) : PatternCollection(db, pFile) {
	// Store top-1 in this array after an iteration
	mNumIterations = config->Read<uint32>("numPatterns");
	mNumPatterns = 0;
	mPatterns = new ItemSet *[mNumIterations];

	// Temporary top-1; for each iteration
	mBestPattern = NULL;
	mBestLambda = DOUBLE_MIN_VALUE;
}
IterativeTop1Collection::~IterativeTop1Collection() {
	delete mBestPattern;
	for(uint32 i=0; i<mNumPatterns; i++)
		delete mPatterns[i];
	delete[] mPatterns;
}
void IterativeTop1Collection::Add(ItemSet *is) {
	if(mBestPattern == NULL) {		// No pattern yet? Use it.
		mBestPattern = is->Clone();
		mBestLambda = is->GetLambda();
		return;
	}

	if(is->GetLambda() > mBestLambda) { // Better than best?
		delete mBestPattern;
		mBestPattern = is->Clone();
		mBestLambda = mBestPattern->GetLambda();
	}
}
void IterativeTop1Collection::NextIteration() {
	if(mNumPatterns == mNumIterations)
		THROW(string("Sorry, that's more than I can handle."));
	mPatterns[mNumPatterns++] = mBestPattern;
	
	mBestPattern = NULL;
	mBestLambda = DOUBLE_MIN_VALUE;
}
void IterativeTop1Collection::Finalize() {
	// Output to .txt-file
	if(mPatternFile!=NULL)
		for(uint32 i=0; i<mNumPatterns; i++)
			mPatternFile->WriteItemSet(mPatterns[i]);
}
void IterativeTop1Collection::WritePatternDetails(const string &filename)	{ 
	FILE *fPatterns;
	if(fopen_s(&fPatterns, filename.c_str(), "w"))
		THROW("Could not write individual pattern stats.");
	fprintf_s(fPatterns, "numNodes;numEdges;interestingness;size;edgeDensity;diameter;triangleDensity;avgDegree;nodes\n");
	for(uint32 i=0; i<mNumPatterns; i++)
		WritePatternWithDetails(fPatterns, mPatterns[i]);
	fclose(fPatterns);
}
void IterativeTop1Collection::WritePatternWithDetails(FILE *file, ItemSet *is) {
	if(file == NULL)
		THROW("No file handle open");
	if(ferror(file))
		THROW("Error writing pattern");

	// Compute
	SubgraphStats stats = is->ComputeSubgraphStats(mDatabase);
	uint32 numNodes = is->GetLength();

	// Write
	fprintf_s(file, "%u;%u;%.03f;%.03f;%.03f;%u;%.03f;%.02f;", numNodes, is->GetNumEdges(), is->GetLambda(), 
						stats.relSize, stats.edgeDensity, stats.diameter, stats.triangleDensity, stats.avgDegree);

	// The set itself
	char buffer[ISCFILE_BUFFER_LENGTH];
	char *bp = buffer;
	uint32 *set = is->GetSet();
	for(uint32 j=0; j<numNodes; j++) {
		_itoa_s(set[j], bp, ISCFILE_BUFFER_LENGTH - (bp - buffer), 10);
		bp += strlen(bp);
		*bp = ',';
		bp++;
	}
	if(numNodes > 0)
		*(bp-1) = '\n';
	fwrite(buffer, sizeof(char), bp - buffer, file);
}

void IterativeTop1Collection::ComputeGlobalStats() {
	if(mNumPatterns==0)
		return;

	mMaxInterest = DOUBLE_MIN_VALUE;
	mMinInterest = DOUBLE_MAX_VALUE;
	mAvgInterest = mAvgSize = mAvgJaccard = 0.0;
	mMaxSize = 0;
	mMinSize = UINT32_MAX_VALUE;
	mMaxJaccard = 0.0;
	mMinJaccard = DOUBLE_MAX_VALUE;
	ItemSet *coverage = ItemSet::CreateEmpty(Uint32ItemSetType, 0);

	for(uint32 i=0; i<mNumPatterns; i++) {
		// Interest
		double l = mPatterns[i]->GetLambda();
		if(l > mMaxInterest)
			mMaxInterest = l;
		if(l < mMinInterest)
			mMinInterest = l;
		mAvgInterest += l;

		// Size
		uint32 s = mPatterns[i]->GetLength();
		if(s > mMaxSize)
			mMaxSize = s;
		if(s < mMinSize)
			mMinSize = s;
		mAvgSize += s;

		// Jaccard
		for(uint32 j=i+1; j<mNumPatterns; j++) {
			double jacc = mPatterns[i]->JaccardDistance(mPatterns[j]);
			if(jacc > mMaxJaccard)
				mMaxJaccard = jacc;
			if(jacc < mMinJaccard)
				mMinJaccard = jacc;
			mAvgJaccard += jacc;
		}

		// Coverage
		coverage->Unite(mPatterns[i]);
	}
	if(mNumPatterns > 1) {
		mAvgInterest /= mNumPatterns;
		mAvgSize /= mNumPatterns;
		mAvgJaccard /= (mNumPatterns * (mNumPatterns-1) / 2);
	} else {
		mMinJaccard = mAvgJaccard = mMaxJaccard = 0.0;
	}

	mCoverage = coverage->GetLength();
	delete coverage;
}
