#include <Config.h>
#include <ArrayUtils.h>

#include "../Bass.h"
#include "../db/Database.h"
#include "../itemstructs/ItemSet.h"
#include "../itemstructs/PatternFile.h"

#include "TopKCollection.h"

TopKCollection::TopKCollection(Config *config, Database *db, PatternFile *pFile) : PatternCollection(db, pFile) {
	mK = config->Read<uint32>("numPatterns");

	mPatterns = new ItemSet *[mK];
	mMinThreshold = DOUBLE_MAX_VALUE;
	mMinIdx = UINT32_MAX_VALUE;

	mSetArray = new uint32[db->GetMaxSetLength()];
}
TopKCollection::TopKCollection(const uint32 k, Database *db) : PatternCollection(db, NULL) {
	mK = k;

	mPatterns = new ItemSet *[mK];
	mMinThreshold = DOUBLE_MAX_VALUE;
	mMinIdx = UINT32_MAX_VALUE;

	mSetArray = new uint32[db->GetMaxSetLength()];
}
TopKCollection::~TopKCollection() {
	for(uint32 i=0; i<mNumPatterns; i++)
		delete mPatterns[i];
	delete[] mPatterns;
	delete[] mSetArray;
}
void TopKCollection::Add(ItemSet *is) {
	if(mNumPatterns < mK) {					// No k patterns found yet? Simply add
		if(is->GetLambda() < mMinThreshold) {
			mMinThreshold = is->GetLambda();
			mMinIdx = (uint32)mNumPatterns;
		}
		mPatterns[mNumPatterns++] = is->Clone();
		return;
	}
											// K patterns -> check if minInterest pattern should be replaced
	if(is->GetLambda() > mMinThreshold) {
		// Replace pattern with minimum interest
		delete mPatterns[mMinIdx];
		mPatterns[mMinIdx] = is->Clone();

		// Scan array for new min interest and idx
		mMinThreshold = mPatterns[0]->GetLambda();
		mMinIdx = 0;
		for(uint32 i=1; i<mK; i++) {
			if(mPatterns[i]->GetLambda() < mMinThreshold) {
				mMinThreshold = mPatterns[i]->GetLambda();
				mMinIdx = i;
			}
		}
	}
}
void TopKCollection::Finalize() {
	// Sort on interestingness (bubble-sort)
	for(uint32 i=1; i<mNumPatterns; i++) {
		for(uint32 j=0; j<mNumPatterns-i; j++) {
			if(mPatterns[j]->GetLambda() < mPatterns[j+1]->GetLambda()) {
				ItemSet *tmp = mPatterns[j];
				mPatterns[j] = mPatterns[j+1];
				mPatterns[j+1] = tmp;
			}
		}
	}

	// Output to .txt-file
	if(mPatternFile!=NULL)
		for(uint32 i=0; i<mNumPatterns; i++)
			mPatternFile->WriteItemSet(mPatterns[i]);
}
void TopKCollection::WritePatternDetails(const string &filename)	{ 
	FILE *fPatterns;
	if(fopen_s(&fPatterns, filename.c_str(), "w"))
		THROW("Could not write individual pattern stats.");
	fprintf_s(fPatterns, "numNodes;numEdges;interestingness;size;edgeDensity;diameter;triangleDensity;nodes\n");
	for(uint32 i=0; i<mNumPatterns; i++)
		WritePatternWithDetails(fPatterns, mPatterns[i]);
	fclose(fPatterns);
}
void TopKCollection::WritePatternWithDetails(FILE *file, ItemSet *is) {
	if(file == NULL)
		THROW("No file handle open");
	if(ferror(file))
		THROW("Error writing pattern");

	// Some detailed statistics of the subgraph
	is->GetValuesIn(mSetArray);
	uint32 numEdges = is->GetNumEdges();
	uint32 numNodes = is->GetLength();
	float networkSize = (float) mDatabase->GetNumRows();

		// Get subgraph
	ItemSet **subgraph = new ItemSet *[numNodes];
	for(uint32 i=0; i<numNodes; i++)
		subgraph[i] = mDatabase->GetRow(mSetArray[i])->Intersection(is);

		// Compute #triangles
	uint32 numTriangles = 0;
	for(uint32 i=0; i<numNodes; i++)
		for(uint32 j=i+1; j<numNodes; j++)
			for(uint32 k=j+1; k<numNodes; k++)
				if(subgraph[i]->IsItemInSet(mSetArray[j]) && subgraph[i]->IsItemInSet(mSetArray[k]) && subgraph[j]->IsItemInSet(mSetArray[k]))
					++numTriangles;

		// Compute diameter
	uint32 diameter = 1;
	bool allConnected = true;
	for(uint32 i=0; i<numNodes; i++)
		if(subgraph[i]->GetLength() < numNodes-1) {
			allConnected = false;
			break;
		}
	while(!allConnected && diameter < 500) { // something must be wrong if we reach that 500..
		++diameter;

		// the following is probably everything-but-smart-and-efficient, and should be re-written when speed becomes an issue.. (should use set operations)
		for(uint32 i=0; i<numNodes; i++)
			for(uint32 j=i+1; j<numNodes; j++)
				for(uint32 k=j+1; k<numNodes; k++) {
					if(subgraph[i]->IsItemInSet(mSetArray[j]) && subgraph[i]->IsItemInSet(mSetArray[k]) && !subgraph[j]->IsItemInSet(mSetArray[k])) {
						subgraph[j]->AddItemToSet(mSetArray[k]);
						subgraph[k]->AddItemToSet(mSetArray[j]);
					}
					if(!subgraph[i]->IsItemInSet(mSetArray[j]) && subgraph[i]->IsItemInSet(mSetArray[k]) && subgraph[j]->IsItemInSet(mSetArray[k])) {
						subgraph[i]->AddItemToSet(mSetArray[j]);
						subgraph[j]->AddItemToSet(mSetArray[i]);
					}
					if(subgraph[i]->IsItemInSet(mSetArray[j]) && !subgraph[i]->IsItemInSet(mSetArray[k]) && subgraph[j]->IsItemInSet(mSetArray[k])) {
						subgraph[i]->AddItemToSet(mSetArray[k]);
						subgraph[k]->AddItemToSet(mSetArray[i]);
					}
				}

		allConnected = true;
		for(uint32 i=0; i<numNodes; i++)
			if(subgraph[i]->GetLength() < numNodes-1) {
				allConnected = false;
				break;
			}
	}

	// Cleanup
	for(uint32 i=0; i<numNodes; i++)
		delete subgraph[i];
	delete[] subgraph;

		// Write
	fprintf_s(file, "%u;%u;%.03f;%.03f;%.03f;%u;%.03f;", numNodes, numEdges, is->GetLambda(), numNodes / networkSize, (float)numEdges / (numNodes * (numNodes-1) / 2),
												diameter, (float)numTriangles / (numNodes * (numNodes-1) * (numNodes-2) / 6));

	// The set itself
	char buffer[ISCFILE_BUFFER_LENGTH];
	char *bp = buffer;
	for(uint32 j=0; j<numNodes; j++) {
		_itoa_s(mSetArray[j], bp, ISCFILE_BUFFER_LENGTH - (bp - buffer), 10);
		bp += strlen(bp);
		*bp = ',';
		bp++;
	}
	if(numNodes > 0)
		*(bp-1) = '\n';
	fwrite(buffer, sizeof(char), bp - buffer, file);
}

void TopKCollection::ComputeGlobalStats() {
	if(mNumPatterns==0)
		return;

	mMaxInterest = mPatterns[0]->GetLambda();
	mMinInterest = mPatterns[mNumPatterns-1]->GetLambda();

	mAvgInterest = mAvgSize = mAvgJaccard = 0.0;
	mMaxSize = 0;
	mMinSize = UINT32_MAX_VALUE;
	mMaxJaccard = 0.0;
	mMinJaccard = DOUBLE_MAX_VALUE;
	ItemSet *coverage = ItemSet::CreateEmpty(Uint32ItemSetType, 0);

	for(uint32 i=0; i<mNumPatterns; i++) {
		// Interest
		mAvgInterest += mPatterns[i]->GetLambda();

		// Size
		uint32 s = mPatterns[i]->GetLength();
		if(s > mMaxSize)
			mMaxSize = s;
		if(s < mMinSize)
			mMinSize = s;
		mAvgSize += s;

		// Jaccard
		for(uint32 j=i+1; j<mNumPatterns; j++) {
			double jacc = mPatterns[i]->JaccardDistance(mPatterns[j]);
			if(jacc > mMaxJaccard)
				mMaxJaccard = jacc;
			if(jacc < mMinJaccard)
				mMinJaccard = jacc;
			mAvgJaccard += jacc;
		}

		// Coverage
		coverage->Unite(mPatterns[i]);
	}
	if(mNumPatterns > 1) {
		mAvgInterest /= mNumPatterns;
		mAvgSize /= mNumPatterns;
		mAvgJaccard /= (mNumPatterns * (mNumPatterns-1) / 2);
	} else {
		mMinJaccard = mAvgJaccard = mMaxJaccard = 0.0;
	}

	mCoverage = coverage->GetLength();
	delete coverage;
}
