#include "../Bass.h"

#include <cstring>

#include <FileUtils.h>
#include <ArrayUtils.h>
#include <RandomUtils.h>
#include <logger/Log.h>

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include "../itemstructs/Uint32ItemSet.h"
#include "DbFile.h"

#include "Database.h"

Database::Database() {
	mDbName = "unknown";
	mRows = NULL;
	mNumRows = 0;
	mNumItems = 0;
	mDataType = NoneItemSetType;
	mFastGenerateDBOccs = false;

	mPath = "";
	mFilename = "";

	// Essentials(tm)
	mAlphabet = NULL;
	mMaxSetLength = 0;

	// Not per-se Essential (tm)
	mAlphabetNumRows = NULL;
	mAlphabetArr = NULL;
	mAlphabetDbOccs = NULL;

	mNumEntityTypes = 0;
	mEntityTypeCounts = NULL;
}

Database::Database(Database *db) {
	mDbName = db->mDbName;
	mNumRows = db->GetNumRows();
	ItemSet **iss = db->GetRows();
	mRows = new ItemSet *[mNumRows];
	for(uint32 i=0; i<mNumRows; i++)
		mRows[i] = iss[i]->Clone();
	mNumItems = db->GetNumItems();
	mDataType = db->GetDataType();

	mPath = "";
	mFilename = "";

	// Essentials(tm)
	mAlphabet = new alphabet(*db->GetAlphabet());
	mMaxSetLength = db->GetMaxSetLength();

	// Not per-se Essential (tm)
	mAlphabetArr = NULL;
	mAlphabetNumRows = NULL;
	if(db->mAlphabetNumRows != NULL) {
		mAlphabetNumRows = new uint32[mAlphabet->size()];
		memcpy(mAlphabetNumRows, db->mAlphabetNumRows, sizeof(uint32) * mAlphabet->size());
	}
	mAlphabetDbOccs = NULL;

	mNumEntityTypes = db->mNumEntityTypes;
	mEntityTypeCounts = NULL;
	if(db->mEntityTypeCounts != NULL) {
		mEntityTypeCounts = new uint32[mNumEntityTypes];
		memcpy(mEntityTypeCounts, db->mEntityTypeCounts, sizeof(uint32) * mNumEntityTypes);
	}
}

Database::Database(ItemSet **itemsets, uint32 numRows, uint32 numEntityTypes /*= 0*/) {
	mDbName = "unknown";
	mRows = itemsets;
	mNumRows = numRows;
	mNumItems = 0;

	mPath = "";
	mFilename = "";

	bool computeNumItems = false;

	if(mNumItems == 0) 
		computeNumItems = true;

	mDataType = numRows > 0 ? itemsets[0]->GetType() : NoneItemSetType;
	mFastGenerateDBOccs = false;

	if(computeNumItems) {
		for(uint32 i=0; i<mNumRows; i++)
			mNumItems += mRows[i]->GetLength();
	}

	// Essentials(tm)
	mAlphabet = NULL;
	mMaxSetLength = 0;

	// Not per-se Essential (tm)
	mAlphabetArr = NULL;
	mAlphabetNumRows = NULL;
	mAlphabetDbOccs = NULL;

	mNumEntityTypes = numEntityTypes;
	mEntityTypeCounts = NULL;
	if(mNumEntityTypes > 0) {
		mEntityTypeCounts = new uint32[mNumEntityTypes];
		for(uint32 i=0; i<mNumEntityTypes; i++)
			mEntityTypeCounts[i] = 0;
		for(uint32 i=0; i<mNumRows; i++)
			;//mEntityTypeCounts[mRows[i]->GetEntityType()]++;
	}
}

Database::~Database() {
	CleanUpStaticMess();
	for(uint32 i=0; i<mNumRows; i++)
		delete mRows[i];
	delete[] mRows;

	delete mAlphabet;
	delete[] mAlphabetNumRows;
	delete[] mAlphabetArr;
	if(mAlphabetDbOccs != NULL) {
		uint32 abSize = (uint32) mAlphabet->size();
		for(uint32 i=0; i<abSize; i++)
			delete[] mAlphabetDbOccs[i];
		delete[] mAlphabetDbOccs;
	}
	if(mEntityTypeCounts != NULL)
		delete[] mEntityTypeCounts;
}

Database* Database::Clone() {
	return new Database(this);
}

void Database::CleanUpStaticMess() {
	// ~~~ if(ItemSet::HasStaticMess())
	if(mNumRows > 0)
		mRows[0]->CleanUpStaticMess();
}

//////////////////////////////////////////////////////////////////////////
// Graph-specific functions
//////////////////////////////////////////////////////////////////////////
void Database::RemoveNodes(ItemSet *nodes) {
	for(uint32 r=0; r<mNumRows; r++) {
		if(nodes->IsItemInSet(r))
			mRows[r]->CleanTheSlate();
		else
			mRows[r]->Remove(nodes);
	}
	mMaxSetLength = 0;
	mNumItems = 0;
	ComputeEssentials();
}

//////////////////////////////////////////////////////////////////////////
//	Alphabet Functions
//////////////////////////////////////////////////////////////////////////

void Database::UseAlphabetFrom(Database *db, bool originalStdLengths) {
	if(mAlphabet) delete mAlphabet;
	delete[] mAlphabetNumRows; mAlphabetNumRows = NULL;
	mAlphabet = new alphabet(*db->GetAlphabet());
	CountAlphabet();
}
void Database::SetAlphabetSize(const uint32 newSize) {
	if(mAlphabet) delete mAlphabet;
	mAlphabet = new alphabet();
	for(uint32 i=0; i<newSize; i++)
		mAlphabet->insert(alphabetEntry(i, 0));
	CountAlphabet();
}
ItemSet** Database::GetAlphabetSets() const {
	ItemSet **abSets = new ItemSet*[mAlphabet->size()];
	alphabet::iterator it = mAlphabet->begin(), end = mAlphabet->end();
	uint32 curIdx = 0;
	uint32 abSize = (uint32) mAlphabet->size();
	for(; it != end; ++it) {
		abSets[curIdx] = ItemSet::CreateSingleton(mDataType, it->first, abSize, it->second, mAlphabetNumRows[curIdx]);
		curIdx++;
	}
	return abSets;
}

//////////////////////////////////////////////////////////////////////////
//	Internal Statistics Functions
//////////////////////////////////////////////////////////////////////////

bool Database::ComputeEssentials() {
	if(!mAlphabet && !this->DetermineAlphabet())
		return false;
	if(mMaxSetLength == 0 || mNumItems == 0)
		this->DetermineMaxSetLength();

	return true;
}

bool Database::DetermineAlphabet() {
	if(mAlphabet)
		return false;
	ECHO(3, printf("Database :: Determining alphabet ... "));
	ItemSet *is;
	uint16 val;
	mAlphabet = new alphabet();
	alphabet *abNR = new alphabet();
	alphabet::iterator iter;
	uint32 cnt;
	mNumItems = 0;

	uint32 maxLen = 0;
	for(uint32 row=0; row<mNumRows; row++)
		if(mRows[row]->GetLength() > maxLen)
			maxLen = mRows[row]->GetLength();
	uint32 *values = new uint32[maxLen];

	for(uint32 row=0; row<mNumRows; row++) {
		is = mRows[row];
		is->GetValuesIn(values);
		cnt = 1;
		for(uint16 col=0; col<is->GetLength(); col++) {
			val = values[col];
			iter = mAlphabet->find(val);
			if(iter == mAlphabet->end()) {
				mAlphabet->insert(alphabetEntry(val, cnt));
				abNR->insert(alphabetEntry(val, 1));
			} else {
				iter->second += cnt;
				iter = abNR->find(val);
				iter->second += 1;
			}
		}
		mNumItems += is->GetLength() * cnt;
	}
	delete[] values;

	delete[] mAlphabetNumRows;
	mAlphabetNumRows = new uint32[(unsigned int) mNumItems];
	uint32 idx = 0;
	for(iter = abNR->begin(); iter != abNR->end(); iter++, idx++)
		mAlphabetNumRows[idx] = iter->second;
	delete abNR;

	return true;
}
void Database::CountAlphabet() {
	ECHO(3,printf(" * Counting alphabet:\tin progress..."));
	uint32 *vals = new uint32[mAlphabet->size()];
	ItemSet *is;
	uint32 len, cnt;
	alphabet::iterator iter;
	for(iter = mAlphabet->begin(); iter != mAlphabet->end(); iter++)
		iter->second = 0;

	for(uint32 row=0; row<mNumRows; row++) {
		is = mRows[row];
		is->GetValuesIn(vals);
		len = is->GetLength();
		cnt = 1;
		for(uint32 col=0; col<len; col++) {
			iter = mAlphabet->find(vals[col]);
			if(iter == mAlphabet->end())
				throw "Item not in alphabet!";
			else
				iter->second += cnt;
		}
	}
	delete[] vals;

	mNumItems = 0;
	for(iter = mAlphabet->begin(); iter != mAlphabet->end(); iter++)
		mNumItems += iter->second;

	ECHO(3,printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\bdone            \n"));
}

//////////////////////////////////////////////////////////////////////////
// Simple Database Operations
//////////////////////////////////////////////////////////////////////////

uint32 Database::GetSupport(ItemSet *is) {
	uint32 freq = 0;
	for(uint32 i=0; i<mNumRows; i++) {
		if(mRows[i]->IsSubset(is))
			++freq;
	}
	return freq;
}

void Database::CalcSupport(ItemSet *is) {
	uint32 freq = 0;
	for(uint32 i=0; i<mNumRows; i++) {
		if(mRows[i]->IsSubset(is))
			++freq;
	}
	//is->SetSupport(freq);
	THROW("Support not supported.")
}
bool Database::IsItemSetInDatabase(ItemSet *is) {
	for(uint32 i=0; i<mNumRows; i++)
		if(mRows[i]->IsSubset(is))
			return true;
	return false;
}


//////////////////////////////////////////////////////////////////////////
// Database Transformation Operations
//////////////////////////////////////////////////////////////////////////

void Database::Transform(string commandlist) {
	// Randomize
	// TranslateValToBit
	// TranslateBitToVal
	// Transpose
	// SplitOn(..)
	// Bin
	// UnBin
	for(uint32 i=0; i<commandlist.length(); i++) {
		string curCommand = commandlist.substr(i,min(commandlist.find_first_of(','),commandlist.length()));
		//curCommand.erase(' ');
	}
}
void Database::RemoveDoubleItems() {
	if(mDataType == Uint32ItemSetType) {
		ECHO(1,	printf(" * Removing double items:\t\tworking on it..."));
		for(uint32 i=0; i<mNumRows; i++)
			((Uint32ItemSet *)mRows[i])->RemoveDoubleItems();
		ECHO(1,	printf("\r * Removing double items:\t\tdone\t\t\t\t\n\n"));
	} else {
		ECHO(1,	printf("\r * Removing double items:\t\tnot needed with bitmask\t\t\n\n"));
	}
}
void Database::RemoveFrequentItems(float frequencyThreshold) {
	ItemSet *zapItems = ItemSet::CreateEmpty(mDataType, (uint32)mAlphabet->size());
	uint32 absFreqThresh = (uint32)ceil(frequencyThreshold * mNumRows);
	for(alphabet::iterator iter = mAlphabet->begin(); iter != mAlphabet->end(); iter++) {
		if(iter->second >= absFreqThresh)
			zapItems->AddItemToSet(iter->first);
	}
	RemoveItems(zapItems);
}
void Database::RemoveItems(ItemSet *zapItems) {
	uint32 *auxArr = zapItems->GetValues();
	uint32 numValues = zapItems->GetLength();
	RemoveItems(auxArr, numValues);
	delete[] auxArr;
}
void Database::RemoveItems(uint32 *zapItems, uint32 numItems) {
	uint32 newNumRows = 0;
	for(uint32 i=0; i<mNumRows; i++) {
		for(uint32 j=0; j<numItems; j++) {
			if(mRows[i]->IsItemInSet(zapItems[j]))
				mRows[i]->RemoveItemFromSet(zapItems[j]);
		}
		if(mRows[i]->GetLength() > 0)
			mRows[newNumRows++] = mRows[i];
		else {
			delete mRows[i];
		}
	}
	mNumRows = newNumRows;
	CountAlphabet();
}


void Database::SortWithinRows() {
	for(uint32 row=0; row<mNumRows; row++)
		mRows[row]->Sort();
}
void Database::SwapRows(uint32 row1, uint32 row2) {
	ItemSet *is = mRows[row1];
	mRows[row1] = mRows[row2];
	mRows[row2] = is;
}

void Database::SwapRandomise(uint64 numSwaps) {
	uint64 num = numSwaps == 0 ? mNumItems : numSwaps;	
	size_t alphLen = mAlphabet->size();
	uint32 len, tries;
	uint32 row1, row2, col1, col2;
	ItemSet *is1, *is2;
	uint32 *vals1 = new uint32[alphLen], *vals2 = new uint32[alphLen];
	uint32 val1, val2;
	bool found;
	RandomUtils::Init();

	for(uint64 i=0; i<num; i++) {
		ECHO(3,	printf(" * Swapping databases:\t\t%d/%d\r", i, num));

		found = false;

		// find swappable items
		while(!found) {
			row2 = row1 = RandomUtils::UniformUint32(mNumRows); // uint32 from [0,mNumRows)
			is1 = mRows[row1];
			is1->GetValuesIn(vals1);
			col1 = RandomUtils::UniformUint32(is1->GetLength());
			val1 = vals1[col1];

			for(tries=0; (row1 == row2 || is2->IsItemInSet(val1)) && tries<mNumRows; tries++) {
				row2 = RandomUtils::UniformUint32(mNumRows); // uint32 from [0,mNumRows)
				is2 = mRows[row2];
				is2->GetValuesIn(vals2);
			}
			if(tries == mNumRows)
				continue;
			len = is2->GetLength();
			for(tries=0; tries<len; tries++) {
				col2 = RandomUtils::UniformUint32(len);
				val2 = vals2[col2];

				if(val1 != val2 && !is1->IsItemInSet(val2)) {
					found = true;
					break;
				}
			}
		}

		// perform swap
		is1->RemoveItemFromSet(val1);
		is1->AddItemToSet(val2);
		is2->RemoveItemFromSet(val2);
		is2->AddItemToSet(val1);
	}
	delete[] vals1;
	delete[] vals2;
}

Database* Database::Subset(uint32 fromRow, uint32 toRow) {
	if(fromRow > toRow)
		return NULL;
	if(toRow > mNumRows)
		return NULL;
	
	uint32 numRows = toRow - fromRow + 1;
	uint32 idx = 0;
	ItemSet **iss = new ItemSet *[numRows];
	for(uint32 r=fromRow; r<=toRow; r++)
		iss[idx++] = mRows[r]->Clone();

	Database *db = new Database(iss, numRows);
	db->SetAlphabet(new alphabet(*mAlphabet));
	db->CountAlphabet();
	db->ComputeEssentials();
	return db;
}

void Database::DetermineMaxSetLength() {
	mMaxSetLength = 0;
	mNumItems = 0;
	for(uint32 i=0; i<mNumRows; i++) {
		uint32 len = mRows[i]->GetLength();
		mNumItems += len;
		if(len > mMaxSetLength)
			mMaxSetLength = len;
	}
}


//////////////////////////////////////////////////////////////////////////
// Database Read/Write Operations
//////////////////////////////////////////////////////////////////////////

void Database::Write(const string &dbName, const string &path, const DbFileType dbType) {
	Database::WriteDatabase(this, dbName, path, dbType);
}

/* Retrieves database from repository directories (gDxecDir, gDataDir, gDBReposDir) */
Database* Database::RetrieveDatabase(const string &dbName, const ItemSetType dataType, const DbFileType dbType) {
	//if(dbName.find_last_of("\\/") != string::npos)
		//throw string("Please provide a database name to retrieve, not a path");

	// - Determine the database filename
	string inDbFileName, inDbRelPath = "";
	if(dbName.find_last_of("\\/") != string::npos) {
		inDbFileName = dbName.substr(dbName.find_last_of("\\/")+1, string::npos);
		inDbRelPath = dbName.substr(0, dbName.find_last_of("\\/")+1);
	} else {
		inDbFileName = dbName;
	}
	size_t splitPos = inDbFileName.find_first_of("[") == string::npos ? inDbFileName.find_last_of(".") : inDbFileName.find_first_of("[");
	string inDbBaseName = inDbFileName.substr(0,splitPos);
	string inDbFileExtension = "";
	if((dbName.find('.') == string::npos)) {
		// - Add right extension, .db as default, 
		inDbFileName += "." + ((dbType != DbFileTypeNone) ? DbFile::TypeToExt(dbType) : "uno");
		inDbFileExtension = (dbType != DbFileTypeNone) ? DbFile::TypeToExt(dbType) : "uno";
	} else {
		inDbFileExtension = dbName.substr(dbName.find_last_of('.')+1);
		if(DbFile::ExtToType(inDbFileExtension) == DbFileTypeNone) { // unknown extension? then probably no extension..
			inDbBaseName = inDbFileName;
			inDbFileName += "." + ((dbType != DbFileTypeNone) ? DbFile::TypeToExt(dbType) : "uno");
			inDbFileExtension = (dbType != DbFileTypeNone) ? DbFile::TypeToExt(dbType) : "uno";
		}
	}

	// - Determine type of file to load
	DbFileType inDbType = dbType == DbFileTypeNone ? DbFile::ExtToType(inDbFileExtension) : dbType;
	if(inDbType == DbFileTypeNone)
		throw string("Unknown database encoding.");

	// - Determine the path to the requested database
	string inDbPath = "";
	if(FileUtils::FileExists(Bass::GetExecDir() + inDbRelPath + inDbFileName))
		inDbPath = string(Bass::GetExecDir()) + inDbRelPath;
	else if(FileUtils::FileExists(Bass::GetWorkingDir() + inDbRelPath + inDbFileName))
		inDbPath = Bass::GetWorkingDir() + inDbRelPath;
	else if(FileUtils::FileExists(Bass::GetExpDir() + inDbRelPath + inDbFileName))
		inDbPath = Bass::GetExpDir() + inDbRelPath;
	else if(FileUtils::FileExists(Bass::GetDataDir() + inDbRelPath + inDbFileName))
		inDbPath = Bass::GetDataDir() + inDbRelPath;
	else if(FileUtils::FileExists(Bass::GetDbReposDir() + inDbRelPath + inDbFileName))
		inDbPath = Bass::GetDbReposDir() + inDbRelPath;
	else if(FileUtils::FileExists(Bass::GetDbReposDir() + inDbBaseName + "/" + inDbFileName))
		inDbPath = Bass::GetDbReposDir() + inDbBaseName + "/";
	else
		throw string("Unable to retrieve " + dbName + " from repository directories.");

	return Database::ReadDatabase(inDbFileName, inDbPath, dataType, inDbType);
}

/* Stores Database* `db` into the DatabaseRepository as a FicDbFile */
void Database::StoreDatabase(Database* const db, const string &dbName) {
	if(dbName.find_last_of("\\/") != string::npos)
		throw string("Please only providy a (file)name to store the database in the repository, not a path");
	if(dbName.length() == 0)
		throw string("Please provide a (file)name to store the database in the repository");
	//if(dbName.find('.') != string::npos && dbName.substr(dbName.find_last_of('.')).compare(DbFile::TypeToExt(FicDbFileType)) != 0)
	//	throw string("Can only store FicDb files into the repository");
	string dbBaseName = dbName.substr(0,dbName.find_first_of("[."));
	string reposPath = Bass::GetDbReposDir();
	if(FileUtils::DirectoryExists(reposPath + dbBaseName))
		reposPath = Bass::GetDbReposDir() + dbBaseName;

	//Database::WriteDatabase(db, dbName, reposPath, DbFileTypeFic);
}


/* Reads database file of `dbName` from `path` into a Database object consisting of rows of ItemSetType `dataType`
	- Preferred use is to split database name (with or without extension) and the path. Alternatively, the path will be extracted from the provided filename.
	- Database (file)name extension is optional. Defaults to ".db"
	- Path defaults to current directory
	- ItemSetType defaults to BM128ItemSet
	- DbFileType defaults to (.db ->) FicDbFileType
*/
Database* Database::ReadDatabase(const string &dbName, const string &path /* optional, defaults to local directory */, const ItemSetType dataType /* optional, BM128ItemSet default */, const DbFileType dbType /* optional, extension-dependent, defaults to FicDbFileType*/) {
	string inDbPath = path;
	string inDbFileName = dbName;
	string inDbFileExtension = "";

	string inDbFullPath = DeterminePathFileExt(inDbPath, inDbFileName, inDbFileExtension, dbType);

	// - Check whether designated file exists
	if(!FileUtils::FileExists(inDbFullPath))
		throw string("Database file does not seem to exist.");

	// - Determine type of file to load
	DbFileType inDbType = dbType == DbFileTypeNone ? DbFile::ExtToType(inDbFileExtension) : dbType;
	if(inDbType == DbFileTypeNone)
		throw string("Unknown database encoding.");

	// - Create appropriate DbFile to load that type of database
	DbFile *dbFile = DbFile::Create(inDbType);

	// - Determine internal datatype to read into
	ItemSetType inDbItemSetType = dataType == NoneItemSetType ? Uint32ItemSetType : dataType;
	dbFile->SetPreferredDataType(inDbItemSetType);

	Database *db = NULL;
	try {	// Memory-leak-protection-try/catch
		db = dbFile->Read(inDbFullPath);
		db->SetName(inDbFileName.substr(0,inDbFileName.find_last_of('.')));
	} catch(...) {	
		delete dbFile;	
		throw;
	}
	delete dbFile;

	if(!db)
		throw string("Could not read input database: " + inDbFullPath);
	return db;
}

// Shallt not fux0r the Database up.
string Database::WriteDatabase(Database * const db, const string &dbName, const string &path, const DbFileType dbType) {
	// - Determine whether `dbName` does not abusively contains a path
	size_t dbNameLastDirSep = dbName.find_last_of("\\/");

	// - Determine `outDbPath`
	string outDbPath = (path.length() == 0 && dbNameLastDirSep != string::npos) ? dbName.substr(0,dbNameLastDirSep) : path;

	//  - Check whether is a valid (slash or backslash-ended) path
	if(outDbPath.length() != 0 && outDbPath[outDbPath.length()-1] != '/' && outDbPath[outDbPath.length()-1] != '\\')
		outDbPath += '/';

	// - Determine `outDbFileName` = dbName + . + extension
	string outDbFileName = (dbNameLastDirSep != string::npos) ? dbName.substr(dbNameLastDirSep+1) : dbName;
	string outDbFileExtension = "";

	// - Determine `outDbType` (preliminary defaulting, in case extension is provided via `dbName`)
	DbFileType outDbType = (dbType == DbFileTypeNone) ? DbFileTypeUno : dbType;

	if(dbName.find('.') == string::npos) {
		//  - No extension provided, so extract from DbType
		outDbFileExtension = DbFile::TypeToExt(outDbType);
		outDbFileName += "." + outDbFileExtension;
	} else {
		// - Wel een punt-stuk in de db(file)name, niet per definitie een extensie
		string posExt = dbName.substr(dbName.find_last_of('.')+1);
		DbFileType posDbType = DbFile::ExtToType(posExt);

		// - Geen aparte extensie meegegeven, wel puntStuk in filename
		if(posDbType == DbFileTypeNone) {
			// - Geen bekende extensie, voeg zelf toe al naar gelang outDbType
			outDbFileExtension = (outDbType != DbFileTypeNone) ? DbFile::TypeToExt(outDbType) : "gdb";
			outDbFileName += "." + outDbFileExtension;
		} else {
			// - Bekende extensie
			if(dbType == DbFileTypeNone) {
				// - gebruik posExt
				outDbType = posDbType;
				outDbFileExtension = posExt;
				// outDbFileName = zit ie al in!
			} else {
				// - gebruik dbType
				outDbFileExtension = (outDbType != DbFileTypeNone) ? DbFile::TypeToExt(outDbType) : "gdb";
				outDbFileName += "." + outDbFileExtension;
			}
		}
	}

	// - Determine `outDbFullPath` = outDbPath + outDbFileName 
	string outDbFullPath = outDbPath + outDbFileName;

	// - Throws error in case unknown dbType.
	DbFile *dbFile = DbFile::Create(outDbType);
	if(!dbFile->Write(db, outDbFullPath)) {
		delete dbFile;
		throw "Error: Could not write database.\n";
	}
	delete dbFile;

	return outDbFullPath;
}

// filename may or may not contain extension
string Database::DeterminePathFileExt(string &path, string &filename, string &extension, const DbFileType dbType) {
	// - Split path from filename (if required, does nothing otherwise)
	FileUtils::SplitPathFromFilename(filename, path);

	// DB files -moeten- een extensie hebben.
	if(filename.find('.') == string::npos) {
		// iig geen extensie in de (file)name
		if(extension.length() == 0) { 
			// - Geen aparte extensie meegegeven, voeg zelf toe al naar gelang dbType
			filename += "." + ((dbType != DbFileTypeNone) ? DbFile::TypeToExt(dbType) : "gdb");
			extension = (dbType != DbFileTypeNone) ? DbFile::TypeToExt(dbType) : "gdb";
		} else {	
			// - Aparte extensie meegegeven, zit per definitie nog niet in filename (nl geen '.' in filename) 
			if(extension[0] == '.')
				extension = extension.substr(1);
			filename += "." + extension;
		}
	} else {
		// - Wel een punt-stuk in de (file)name, niet per definitie een extensie
		string posExt = filename.substr(filename.find_last_of('.')+1);
		if(extension.length() == 0) {	
			// - Geen aparte extensie meegegeven, wel puntStuk in filename
			if(DbFile::ExtToType(posExt) == DbFileTypeNone) {
				// - Geen bekende extensie, voeg zelf toe al naar gelang dbType
				filename += "." + ((dbType != DbFileTypeNone) ? DbFile::TypeToExt(dbType) : "gdb");
				extension = (dbType != DbFileTypeNone) ? DbFile::TypeToExt(dbType) : "gdb";
			} else {
				// - Bekende extensie, zal 'm wel zijn dan
				extension = posExt;
				if(extension[0] == '.')
					extension = extension.substr(1);
			}
		} else {	
			// - Wel aparte extensie meegegeven
			if(extension[0] == '.')
				extension = extension.substr(1);
			if(posExt.compare(extension) != 0) {
				// - Alleen als anders dan opgegeven, voeg 'm maar toe
				filename += "." + extension;
			}	
		}
	}


	// - Determine `inDbFullPath`
	string fullPath = path + filename;

	// - Check whether designated file exists
	/*if(!FileUtils::FileExists(fullPath))
		throw string("Database file does not seem to exist.");*/

	return fullPath;
}

Database* Database::GenerateErdosRenyi(const uint32 numNodes, const double density) {
	// Init structure
	ItemSet **temp = new ItemSet *[numNodes];
	ItemSet **rows = new ItemSet *[numNodes];
	for (uint32 r = 0; r < numNodes; r++) {
		temp[r] = ItemSet::CreateEmpty(Uint32ItemSetType, numNodes);
		rows[r] = ItemSet::CreateEmpty(Uint32ItemSetType, numNodes);
	}

	// Create randomized structure
	for (uint32 r = 1; r < numNodes; r++) {
		for (uint32 s = 0; s < r; s++) {
			if (RandomUtils::UniformBool(density)) {
				temp[r]->AddItemToSet(s);
				temp[s]->AddItemToSet(r);
			}
		}
	}

	// Create degree mapping
	uint32 *degrees = new uint32[numNodes];
	for (uint32 r = 0; r < numNodes; r++)
		degrees[r] = temp[r]->GetLength();
	uint32 *mapping = RandomUtils::GenerateOrderedList(0, numNodes);
	ArrayUtils::BSortArrayDesc<uint32>(degrees, numNodes, mapping);

	// Create the right data
	for (uint32 r = 0; r < numNodes; r++) {
		Uint32ItemSet *is = (Uint32ItemSet*)(temp[mapping[r]]);
		uint32 *vals = is->GetSet();
		for (uint32 i = 0; i < is->GetLength(); i++) {
			rows[r]->AddItemToSet(mapping[vals[i]]);
		}
	}

	// Cleanup
	for (uint32 r = 0; r < numNodes; r++)
		delete temp[r];
	delete[] temp;
	delete[] degrees;
	delete[] mapping;

	// Create the database
	Database *db = new Database(rows, numNodes);
	db->ComputeEssentials();
	return db;
}
