#ifndef __DATABASE_H
#define __DATABASE_H

#include "../Bass.h"
#include "../itemstructs/ItemSet.h"
#include "DbFile.h"

class ItemTranslator;

enum DatabaseType {
	DatabasePlain,
	DatabaseClassed,
};

/*
	Rows are the physical, in memory, ItemSets. These can be grouped, and thus have counts >1. 
	Transactions are the virtual, ungrouped, ItemSets.
*/

class BASS_API Database {
public:
	Database();
	Database(Database *database);
	Database(ItemSet **itemsets, uint32 numRows, uint32 numEntityTypes = 0);
	virtual ~Database();
	virtual Database*	Clone();

	void				Write(const string &dbName, const string &path = "", const DbFileType dbType = DbFileTypeNone);

	void				CleanUpStaticMess();
	virtual DatabaseType GetType() { return DatabasePlain; }

	ItemSet*			GetRow(uint32 row)			{ return mRows[row]; }
	ItemSet**			GetRows()					{ return mRows; }
	ItemSet**			GetRows() const				{ return mRows; }

	void				SetRows(ItemSet **rows)		{ mRows = rows; }
	bool				IsItemSetInDatabase(ItemSet *is);

	uint32				GetNumRows() const			{ return mNumRows; }
	void				SetNumRows(uint32 nr)		{ mNumRows = nr; }

	ItemSetType 		GetDataType() const			{ return mDataType; }
	void				SetDataType(ItemSetType ist){ mDataType = ist; }

	string				GetName() const				{ return mDbName; }
	void				SetName(const string &name) { mDbName = name; }

	uint32				GetMaxSetLength() const		{ return mMaxSetLength; }
	void				SetMaxSetLength(uint32 msl)	{ mMaxSetLength = msl; }

	uint32*				GetAlphabetArr() const		{ return mAlphabetArr; }
	alphabet*			GetAlphabet() const		{ return mAlphabet; }
	void				SetAlphabet(alphabet *a)	{ mAlphabet = a; }
	void				UseAlphabetFrom(Database *db, bool originalStdLengths = false);
	size_t				GetAlphabetSize() const { return mAlphabet->size(); }
	void				SetAlphabetSize(const uint32 newSize); // watch out with this one!

	bool				HasMultipleEntityTypes()	{ return mNumEntityTypes > 0; }
	uint32				GetNumEntityTypes()			{ return mNumEntityTypes; }

	// Creates a caller-owned array of caller-owned ItemSets corresponding to the alphabet elements of the database
	ItemSet**			GetAlphabetSets() const;

	uint32*				GetAlphabetNumRows() const	{ return mAlphabetNumRows; }
	void				SetAlphabetNumRows(uint32 *abRo) { mAlphabetNumRows = abRo; }

	uint64				GetNumItems() const			{ return mNumItems; }
	void				SetNumItems(uint64 ni)		{ mNumItems = ni; }

	//////////////////////////////////////////////////////////////////////////
	// Graph-specific functions
	//////////////////////////////////////////////////////////////////////////
	void				RemoveNodes(ItemSet *nodes);

	//////////////////////////////////////////////////////////////////////////
	// Path/Filename functions
	//////////////////////////////////////////////////////////////////////////
	bool				HasFilename() const			{ return mFilename.compare("") != 0; }
	string				GetFilename() const			{ return mFilename; }
	bool				HasPath() const				{ return mPath.compare("") != 0; }
	string				GetPath() const				{ return mPath; }
	bool				HasFullPath() const			{ return HasPath() && HasFilename(); }
	string				GetFullPath() const			{ return mPath + mFilename; }

	//////////////////////////////////////////////////////////////////////////
	// DB Analysis/Statistical/Ladiblabla Operations
	//////////////////////////////////////////////////////////////////////////

	// calculates support and numOccs of given item set and returns this
	uint32				GetSupport(ItemSet *is); 
	// calculates support of given item sets and sets this
	void				CalcSupport(ItemSet *is); 
	
	bool				ComputeEssentials();
	void				CountAlphabet();

	//////////////////////////////////////////////////////////////////////////
	// DB Transformation Operations
	//////////////////////////////////////////////////////////////////////////
	void				Transform(string commandlist);
	void				SortWithinRows();
	void				RemoveDoubleItems(); // uint16, only after sorting within rows!
	void				RemoveFrequentItems(float frequencyThreshold);
	void				RemoveItems(ItemSet *zapItems);
	void				RemoveItems(uint32 *zapItems, uint32 numItems);

	Database*			Subset(uint32 fromRow, uint32 toRow);

	//////////////////////////////////////////////////////////////////////////
	// Read/Write
	//////////////////////////////////////////////////////////////////////////
	static Database*	ReadDatabase (const string &dbName, const string &path = "", const ItemSetType dataType = NoneItemSetType, const DbFileType dbType = DbFileTypeNone);
	static string		WriteDatabase (Database* const db, const string &dbName, const string &path = "", const DbFileType dbType = DbFileTypeNone);

	static Database*	RetrieveDatabase(const string &dbName, const ItemSetType dataType = NoneItemSetType, const DbFileType dbType = DbFileTypeNone);
	static void			StoreDatabase(Database* const db, const string &dbName);

	// Returns full path (path + filename)
	static string		DeterminePathFileExt(string &path, string &filename, string &extension, const DbFileType dbType = DbFileTypeNone);

	// ergens nodig.
	virtual void		SwapRows(uint32 row1, uint32 row2);
	// Swaps numSwaps (defaults to mNumItems, if set to 0) ones in the database, keeping row and column counts the same.
	virtual void		SwapRandomise(uint64 numSwaps = 0);

	///////////////////////////////////////////////////////////////////////////
	// Generate
	///////////////////////////////////////////////////////////////////////////
	static Database*	GenerateErdosRenyi(const uint32 numNodes, const double density);

protected:
	//////////////////////////////////////////////////////////////////////////
	// Housekeeping
	//////////////////////////////////////////////////////////////////////////
	void				SetPath(string p) { mPath = p; }
	void				SetFilename(string f) { mFilename = f; }
	bool				DetermineAlphabet();
	void				DetermineMaxSetLength();


	ItemSet		**mRows;
	ItemSetType	mDataType;
	uint32		mNumRows;			// -- Number of (physical) rows in database
	uint64		mNumItems;
	uint32		mMaxSetLength;
	bool		mFastGenerateDBOccs;

	uint32		mNumEntityTypes;
	uint32		*mEntityTypeCounts;

	string		mDbName;			// de naam van de database. later ook nog een content-code.
	string		mFilename;			// filename of the database, without path.
	string		mPath;				// path of database, without filename.

	alphabet	*mAlphabet;
	size_t		mAlphabetSize;
	uint32		*mAlphabetArr;
	uint32		*mAlphabetNumRows;	// lengte mAlphabet, bevat numRows per ab elem
	uint32		**mAlphabetDbOccs;	// tid list per alphabet item, generated at EnableFastDBOccurences()

	// FimiDbFile is our homie, as it might need to access SortWithinRows stuff
	friend class UnoDbFile;
};

#endif // __DATABASE_H
