#ifndef __ITEMSET_H
#define __ITEMSET_H

// For InterlockedIncrement and InterlockedDecrement
#if defined (_WINDOWS)
	#define WIN32_LEAN_AND_MEAN
	#include <windows.h>
#endif

#include "../BassApi.h"

using namespace std;

struct SubgraphStats {
	uint32	diameter;
	float	relSize;
	float	edgeDensity;
	float	triangleDensity;
	float	avgDegree;
};

enum ItemSetType {
	NoneItemSetType = 0,
	Uint32ItemSetType,
};
//Uint16ItemSetType,
//	BAI32ItemSetType,
//	BM128ItemSetType

enum ItemSetOrder {
	ItemSetOrderNone = 0,
};

// For some reason GCC fails to compile if IscOrderType is defined in isc/ItemSetCollection.h
enum IscOrderType {
	NoneIscOrder=0,			// 0	u
	UnknownIscOrder,		// 1	u
	LengthDescIscOrder,		// 2	d
	LengthAscIscOrder,		// 3	a
	AreaLengthIscOrder,		// 4	al
	AreaSupportIscOrder,	// 5	as
	AreaSquareIscOrder,		// 6	aq
	RandomIscOrder,			// 7	r
	LengthAFreqDIscOrder,	// 8	z
	LengthDFreqDIscOrder,	// 9	x
	LengthAFreqAIscOrder,	// 10	q
	JTestIscOrder,			// 11	j
	JTestRevIscOrder,		// 12	jr
	LexAscIscOrder,			// 13	la
	LexDescIscOrder,		// 14	ld
	IdAscLexAscIscOrder,	// 15
	UsageDescIscOrder,		// 16
};

class Database;
class ItemTranslator;
class ItemSet;

typedef std::list<ItemSet*> islist;
typedef int8 (ItemSet::*ItemSetCompFunc)(const ItemSet *b) const;

class BASS_API ItemSet {
public:
	ItemSet();
	ItemSet(const ItemSet& is);
	virtual ~ItemSet();
	virtual void		CleanUpStaticMess() {};
	virtual ItemSet*	Clone() const =0;

	// Calculates (approximately) how much memory the ItemSet uses
	virtual uint32		GetMemUsage() const =0;

	// GetLength is not const as the bitmap-implementations may require a recount
	virtual uint32		GetLength() =0;
	void __inline		DoNothing() const			{ 	}

	virtual void		AddItemToSet(uint32 val) { throw "Not implemented for this type of itemset."; }
	virtual void		RemoveItemFromSet(uint32 val) { throw "Not implemented for this type of itemset."; }
	virtual bool		IsItemInSet(uint32 val) const =0;
	// Returns the last item of the set this ItemSet was initialised with. 
	// Only use if length>0, as that goes unchecked here.
	virtual const uint32	GetLastItem() const =0;

	// Fills the set with 0's from [o,max>
	virtual void		FillHerUp(uint32 max)=0;
	virtual void		CleanTheSlate()=0;

	// returns the anwser whether 'a' and 'b' contain exactly the same items.
	virtual bool		Equals(ItemSet *is) const =0;

	/* returns the answer to the question whether 'b' intersects with 'a'.
		a->Intersects(b), a = {1,2}, b = {1} -> true
		a->Intersects(b), a = {2}, b = {1} -> false
	*/
	virtual bool		Intersects(ItemSet *is) const =0;
	virtual uint32		IntersectionSize(ItemSet *is) const =0;

	virtual float		JaccardDistance(ItemSet *is) const =0;

	/* returns the answer to the question whether 'b' is a subset of 'a'.
		a->IsSubset(b), a = {1,2}, b = {1} -> true
		a->IsSubset(b), a = {1}, b = {1,2} -> false
	*/
	virtual bool		IsSubset(ItemSet *is) const =0;

	virtual uint32		CountItemsGEQThan(const uint32 n) const =0;

	// zaps the items of is from this
	virtual void		Remove(ItemSet *is) =0;

	virtual ItemSet*	Remainder(ItemSet *is) const;
	virtual ItemSet*	Union(ItemSet *is) const =0;
	virtual ItemSet*	Intersection(ItemSet *is) const =0;

	// unites items of is to this
	virtual void		Unite(ItemSet *is) =0;

	virtual void		Sort() =0;	// may have empty implementation when not applicable for datatype

	// returns ownership of an uint16 array with therein the values of the item set
	virtual uint32*		GetValues() const =0;
	// returns values of the item set in the provided uint32 values array
	virtual void		GetValuesIn(uint32* vals) const =0;
	// increment the values of the item set in the provided uint32 values array
	virtual void		IncrementValues(uint32* vals, int32 cnt) const =0;

	virtual uint32*		GetSet() const =0;

	//////////////// SUBGRAPH-SPECIFIC METHODS /////////////////

	uint32				GetNumEdges() const			{ return mNumEdges; }
	void				SetNumEdges(const uint32 f)	{ mNumEdges = f; }

	double __inline		GetEdgeProbability() const { return mEdgeProbability; }
	void				SetEdgeProbability(const double ep) { mEdgeProbability = ep; }

	double __inline		GetLambda() const { return mLambda; }
	void				SetLambda(const double l) { mLambda = l; }

	double __inline		GetUpperBound() const { return mUpperBound; }
	void				SetUpperBound(const double ub) { mUpperBound = ub; }

	virtual SubgraphStats ComputeSubgraphStats(Database *db) =0;

	///// Rest ////

	virtual string		ToString(bool printCount=true, bool printFreq=false) =0;

	virtual ItemSetType GetType() const =0;

	uint32 __inline		GetId() { return mId; }
	void				SetId(uint32 id) { mId = id; }

	static ItemSet*		Create(ItemSetType type, const uint32 *set, uint32 setlen, uint32 abetlen, uint32 numEdges=0, double lambda=0);
	static ItemSet*		CreateSingleton(ItemSetType type, uint32 elem, uint32 abetlen, uint32 numEdges=0, double lambda=0);
	static ItemSet*		CreateRange(ItemSetType type, uint32 from, uint32 to, uint32 abetlen, uint32 numEdges=0, double lambda=0);
	static ItemSet*		CreateEmpty(ItemSetType type, size_t abetlen);

	// Calculates the approximate memory usage of an ItemSet of `type` with an alphabet of `abLen` and `setLen` items
	static uint32		MemUsage(const ItemSetType type, const uint32 abLen, const uint32 setLen);
	// Calculates the approximate memory usage of a pattern (ItemSet) based on the given database
	static uint32		MemUsage(Database *db);

	//////////////////////////////////////////////////////////////////////////
	// ItemSetType Functions
	//////////////////////////////////////////////////////////////////////////
	static ItemSetType	StringToType(const string desc);
	static string		TypeToString(const ItemSetType type);
	static string		TypeToFullName(const ItemSetType type);

protected:
	uint32	mNumEdges;				// number of edges within subgraph
	double	mEdgeProbability;		// average probability for the existence of an edge within the subgraph, based on background distribution
	double	mLambda;				// interestingness value
	double	mUpperBound;			// upper-bound on interestingness
	uint32	mId;
};

#endif // __ITEMSET_H
