
#include "../Bass.h"
#include "../db/Database.h"

#include "Uint32ItemSet.h"

Uint32ItemSet::Uint32ItemSet(const Uint32ItemSet *iset) : ItemSet(*iset) {
	mLength = iset->mLength;
	mSet = new uint32[mLength];
	memcpy(mSet,iset->mSet, sizeof(uint32) * mLength);
	mLambda = iset->mLambda;
	mUpperBound = DOUBLE_MAX_VALUE;
}
Uint32ItemSet::Uint32ItemSet() {
	mLambda = 0;
	mLength = 0;
	mSet = NULL;
	mUpperBound = DOUBLE_MAX_VALUE;
}

Uint32ItemSet::Uint32ItemSet(uint32 elem, uint32 numEdges, double lambda) {
	mNumEdges = numEdges;
	mLambda = lambda;
	mLength = 1;
	mSet = new uint32[1];
	mSet[0] = elem;
	mUpperBound = DOUBLE_MAX_VALUE;
}
Uint32ItemSet::Uint32ItemSet(uint32 from, uint32 to, uint32 numEdges, double lambda) {
	mLambda = lambda;
	mNumEdges = numEdges;
	mLength = to - from;
	mSet = new uint32[mLength];
	for(uint32 i=0; i<mLength; i++)
		mSet[i] = from + i;
	mUpperBound = DOUBLE_MAX_VALUE;
}
Uint32ItemSet::Uint32ItemSet(const uint32 *set, uint32 setlen, uint32 numEdges, double lambda) {
	mLambda = lambda;
	mNumEdges = numEdges;
	mLength = setlen;
	mSet = new uint32[mLength];
	memcpy(mSet, set, sizeof(uint32) * mLength);
	mUpperBound = DOUBLE_MAX_VALUE;
}

Uint32ItemSet::~Uint32ItemSet() {
	delete[] mSet;
}

Uint32ItemSet *Uint32ItemSet::Clone() const {
	return new Uint32ItemSet(this);
}

uint32 Uint32ItemSet::GetMemUsage() const {
	return sizeof(Uint32ItemSet) + 32 + (mLength * sizeof(uint32));
}
uint32 Uint32ItemSet::MemUsage(const uint32 setLen) {
	return sizeof(Uint32ItemSet) + 32 + (setLen * sizeof(uint32));
}

void Uint32ItemSet::AddItemToSet(uint32 val) {
	if(this->IsItemInSet(val))
		return;
	++mLength;
	uint32 *set = new uint32[mLength];
	uint32 idx = 0;
	uint32 i = 0;
	for(i=0; i<mLength-1; i++)
		if(val > mSet[i])
			set[i] = mSet[i];
		else
			break;
	set[i++] = val;
	for(; i<mLength; i++)
		set[i] = mSet[i-1];
	delete[] mSet;
	mSet = set;
}
void Uint32ItemSet::RemoveItemFromSet(uint32 val) { 
	if(!this->IsItemInSet(val))
		return;
	--mLength;
	uint32 *set = new uint32[mLength];
	uint32 idx = 0;
	for(uint32 i=0; i<=mLength; i++)
		if(mSet[i] != val)
			set[idx++] = mSet[i];
	delete[] mSet;
	mSet = set;
}
void Uint32ItemSet::RemoveDoubleItems() {
	if(mLength <= 1) return;

	uint32 *newValues = new uint32[mLength];
	uint32 newLen = 1;
	newValues[0] = mSet[0];
	for(uint32 i=1; i<mLength; i++)
		if(mSet[i] != newValues[newLen-1])
			newValues[newLen++] = mSet[i];

	if(newLen != mLength) {
		delete[] mSet;
		mLength = newLen;
		mSet = new uint32[mLength];
		memcpy(mSet, newValues, sizeof(uint32) * mLength);
	}
	delete[] newValues;
}

bool Uint32ItemSet::IsItemInSet(uint32 val) const {
	for(uint32 i=0; i<mLength; i++)
		if(mSet[i] >= val)
			return mSet[i] == val;
	return false;
}

int8 Uint32ItemSet::CmpLexicographically(ItemSet *is)  {
	uint32 isLen = is->GetLength();
	uint32 minLen = min(mLength, isLen);
	uint32 *set = ((Uint32ItemSet*) is)->GetSet();
	for(uint32 i=0; i<minLen; i++)
		if(mSet[i] < set[i])
			return -1;
		else if(mSet[i] > set[i])
			return 1;
	if(mLength == isLen)
		return 0;
	return mLength < isLen ? -1 : 1;
}

bool Uint32ItemSet::Equals(ItemSet *is) const {
	if(mLength != is->GetLength())
		return false;

	uint32 *set = ((Uint32ItemSet*) is)->GetSet();
	for(uint32 i=0; i<mLength; i++)
		if(mSet[i] != set[i])
			return false;

	return true;
}

bool Uint32ItemSet::Intersects(ItemSet *is) const {
	uint32 curA=0, numA = mLength;
	uint32 curB=0, numB = is->GetLength();
	uint32 *set = ((Uint32ItemSet*)is)->GetSet();

	while(curA < numA && curB < numB) {
		if(mSet[curA] > set[curB]) {
			curB++;
		} else if(mSet[curA] < set[curB]) {
			curA++;
		} else { 						// gelijk, dus intersects
			return true;
		}
	} return false;
}
uint32 Uint32ItemSet::IntersectionSize(ItemSet *is) const {
	uint32 num = 0;
	uint32 curA=0, numA = mLength;
	uint32 curB=0, numB = is->GetLength();
	uint32 *set = ((Uint32ItemSet*)is)->GetSet();

	while(curA < numA && curB < numB) {
		if(mSet[curA] > set[curB])
			curB++;
		else if(mSet[curA] < set[curB])
			curA++;
		else {						// gelijk, voeg toe, up a&b
			num++; curA++; curB++;
		}
	}
	return num;
}
float Uint32ItemSet::JaccardDistance(ItemSet *is) const {
	uint32 num = 0;
	uint32 curA=0, numA = mLength;
	uint32 curB=0, numB = is->GetLength();
	uint32 *set = ((Uint32ItemSet*)is)->GetSet();

	while(curA < numA && curB < numB) {
		if(mSet[curA] > set[curB])
			curB++;
		else if(mSet[curA] < set[curB])
			curA++;
		else {						// gelijk, voeg toe, up a&b
			num++; curA++; curB++;
		}
	}
	return 1.0f - num / (float)(numA + numB - num);
}

/* returns the answer to the question whether 'b' is a subset of 'a'.
	 a->IsSubset(b), a = {1,2}, b = {1} -> true
	 a->IsSubset(b), a = {1}, b = {1,2} -> false
*/
bool Uint32ItemSet::IsSubset(ItemSet *is) const {
	uint32 num = 0;
	uint32 curA=0, numA = mLength;
	uint32 curB=0, numB = is->GetLength();
	if(numB > numA) return false;

	uint32 *bSet = ((Uint32ItemSet*)is)->GetSet();
	while(curA < numA && curB < numB) {
		if(mSet[curA] > bSet[curB]) {
			curB++;
		} else if(mSet[curA] < bSet[curB]) {
			curA++;
		} else {
			num++; curA++; curB++;
		}
	}
	return (num == numB);
}

uint32 Uint32ItemSet::CountItemsGEQThan(const uint32 n) const {
	uint32 count = 0;
	uint32 idx = mLength - 1;
	while(mSet[idx] >= n) {
		++count;
		if(idx == 0)
			break;
		--idx;
	}
	return count;
}

/* Removes elements in 'is' from our own set */
void Uint32ItemSet::Remove(ItemSet *is) {
	uint32 num = 0;
	uint32 curA=0, numA = mLength;
	uint32 curB=0, numB = is->GetLength();
	uint32 *bSet = ((Uint32ItemSet*)is)->GetSet();

	// bepaal overlap
	while(curA < numA && curB < numB) {
		if(mSet[curA] > bSet[curB]) {
			curB++;
		} else if(mSet[curA] < bSet[curB]) {
			curA++;
		} else {						// gelijk, dus zap
			num++; curA++; curB++;
		}
	}

	if(num > 0) {			// overlap is groter dan 1, dus we moeten items zappen.
		uint32 cur = 0;
		curA=0; curB=0;
		uint32 *rem = new uint32[mLength - num];

		while(curA < numA && curB < numB) {
			if(mSet[curA] > bSet[curB]) {
				curB++;
			} else if(mSet[curA] < bSet[curB]) {
				rem[cur++] = mSet[curA++];
			} else {	// gelijk
				curA++; curB++;
			}
		}
		if(curA < numA)
			for(; curA<numA; curA++)
				rem[cur++] = mSet[curA];

		delete[] mSet;
		mSet = rem;
		mLength = cur;
	}
}

ItemSet* Uint32ItemSet::Intersection(ItemSet *is) const  {
	uint32 num = 0;
	uint32 curA=0, numA = mLength;
	uint32 curB=0, numB = is->GetLength();
	uint32 *set = ((Uint32ItemSet*)is)->GetSet();

	while(curA < numA && curB < numB) {
		if(mSet[curA] > set[curB])
			curB++;
		else if(mSet[curA] < set[curB])
			curA++;
		else {						// gelijk, voeg toe, up a&b
			num++; curA++; curB++;
		}
	}

	uint32 cur = 0;
	curA=0; curB=0;
	uint32 *inter = new uint32[num];
	while(curA < numA && curB < numB) {
		if(mSet[curA] > set[curB])
			curB++;
		else if(mSet[curA] < set[curB])
			curA++;
		else {	// gelijk
			inter[cur++] = set[curB++];
			curA++;
		}
	}

	Uint32ItemSet *isi = new Uint32ItemSet();
	isi->SetSet(inter,num);
	return isi;
}

ItemSet *Uint32ItemSet::Union(ItemSet *is) const  {
	uint32 num = 0;
	uint32 curA=0, numA = mLength;
	uint32 curB=0, numB = is->GetLength();
	uint32 *set = ((Uint32ItemSet*)is)->GetSet();

	while(curA < numA && curB < numB) {
		if(mSet[curA] > set[curB]) {
			num++; curB++;							// voeg B toe
		} else if(mSet[curA] < set[curB]) {
			num++; curA++;							// voeg A toe
		} else {
			num++; curA++; curB++;					// voeg A of B toe
		}
	}
	if(curA < numA)
		num += numA - curA;
	if(curB < numB)
		num += numB - curB;

	uint32 cur = 0; curA=0; curB=0;
	uint32 *uni = new uint32[num];
	while(curA < numA && curB < numB) {
		if(mSet[curA] > set[curB]) {
			uni[cur] = set[curB];
			cur++;	curB++;
		} else if(mSet[curA] < set[curB]) {
			uni[cur] = mSet[curA];
			cur++;	curA++;
		} else {
			uni[cur] = mSet[curA];
			cur++;	curA++;	curB++;
		}
	}
	if(curA < numA)
		for(uint32 i=curA; i<numA; i++)
			uni[cur++] = mSet[i];
	if(curB < numB)
		for(uint32 i=curB; i<numB; i++)
			uni[cur++] = set[i];

	Uint32ItemSet *isi = new Uint32ItemSet();
	isi->SetSet(uni,num);
	return isi;
}

void Uint32ItemSet::Unite(ItemSet *is) {
	uint32 num = 0;
	uint32 curA=0, numA = mLength;
	uint32 curB=0, numB = is->GetLength();
	uint32 *set = ((Uint32ItemSet*)is)->GetSet();

	while(curA < numA && curB < numB) {
		if(mSet[curA] > set[curB]) {
			num++; curB++;							// voeg B toe
		} else if(mSet[curA] < set[curB]) {
			num++; curA++;							// voeg A toe
		} else {
			num++; curA++; curB++;					// voeg 1x toe
		}
	}
	if(curA < numA)
		num += numA - curA;
	if(curB < numB)
		num += numB - curB;

	uint32 cur = 0; curA=0; curB=0;
	uint32 *uni = new uint32[num];
	while(curA < numA && curB < numB) {
		if(mSet[curA] > set[curB]) {
			uni[cur] = set[curB];
			cur++;	curB++;
		} else if(mSet[curA] < set[curB]) {
			uni[cur] = mSet[curA];
			cur++;	curA++;
		} else {
			uni[cur] = mSet[curA];
			cur++;	curA++;	curB++;
		}
	}
	if(curA < numA)
		for(uint32 i=curA; i<numA; i++)
			uni[cur++] = mSet[i];
	if(curB < numB)
		for(uint32 i=curB; i<numB; i++)
			uni[cur++] = set[i];

	delete mSet;

	mSet = uni;
	mLength = num;
}


void Uint32ItemSet::Sort() {
	this->QSortAscending(0, mLength-1);
}

void Uint32ItemSet::QSortAscending(int first, int last) {
	if(last > first +1) {
		int pivot = first;
		int left = first + 1;
		int right = last;

		while (left != right-1) {
			if(mSet[left] <= mSet[pivot])
				left++;
			else 
				this->SwapItems(left, right--);
		}
		if(mSet[left] <= mSet[pivot] &&
			mSet[right] <= mSet[pivot])
			left=right+1;
		else if(mSet[left] <= mSet[pivot] && 
			mSet[right] > mSet[pivot]) {
				left++;
				right--;
		} else if (mSet[left] > mSet[pivot] &&
			mSet[right] <= mSet[pivot])
			this->SwapItems(left++, right--);
		else
			right=left-1;

		this->SwapItems(right--, first);
		this->QSortAscending(first, right);
		this->QSortAscending(left, last);
	} else if(last == first+1) {
		if(mSet[last] < mSet[first])
			this->SwapItems(last,first);
	}
}
string Uint32ItemSet::ToString(bool printNums, bool printLambda) {
	string s = "";
	char *buffer = new char[40];
	if(mLength > 0) {
		for(uint32 i=0; i<mLength-1; i++) {
			sprintf_s(buffer, 40, "%d,", mSet[i]);
			s.append(buffer);
		}
		sprintf_s(buffer, 40, "%d", mSet[mLength-1]);
		s.append(buffer);
	}
	if(printNums || printLambda) {
		if(printNums && !printLambda) {
			sprintf_s(buffer, 40, " (%d,%d)", mNumEdges, mLength);
		} else if(!printNums && printLambda) {
			sprintf_s(buffer, 40, " (%f)", mLambda);
		} else
			sprintf_s(buffer, 40, " (%d,%d,%f)", mNumEdges, mLength, mLambda);
		s.append(buffer);
	}
	delete[] buffer;
	return s;
}

uint32* Uint32ItemSet::GetValues() const {
	uint32 *vals = new uint32[mLength];
	memcpy(vals, mSet, mLength * sizeof(uint32));
	return vals;
}
void Uint32ItemSet::GetValuesIn(uint32* vals) const {
	memcpy(vals,mSet,sizeof(uint32) * mLength);
}
void Uint32ItemSet::IncrementValues(uint32* vals, int32 cnt) const {
	for(uint32 i=0; i<mLength; i++)
		vals[mSet[i]] += cnt;
}


void Uint32ItemSet::FillHerUp(uint32 max) {
	if(mLength != max) {
		delete[] mSet;
		mSet = new uint32[max];		
		mLength = max;
	}
	for(uint32 i=0; i<max; i++) {
		mSet[i] = i;
	}
}

void Uint32ItemSet::CleanTheSlate() {
	mLength = 0;
}

SubgraphStats Uint32ItemSet::ComputeSubgraphStats(Database *db) {
	SubgraphStats stats;

	// Relative size
	stats.relSize = mLength / (float)(db->GetNumRows());

	// Get subgraph
	ItemSet **subgraph = new ItemSet *[mLength];
	for(uint32 i=0; i<mLength; i++)
		subgraph[i] = this->Intersection(db->GetRow(mSet[i]));

	// Compute #triangles
	uint32 numTriangles = 0;
	for(uint32 i=0; i<mLength; i++)
		for(uint32 j=i+1; j<mLength; j++)
			for(uint32 k=j+1; k<mLength; k++)
				if(subgraph[i]->IsItemInSet(mSet[j]) && subgraph[i]->IsItemInSet(mSet[k]) && subgraph[j]->IsItemInSet(mSet[k]))
					++numTriangles;
	stats.triangleDensity = (float)numTriangles / (mLength * (mLength-1) * (mLength-2) / 6);

	// Compute diameter
	stats.diameter = 1;
	bool allConnected = true;
	for(uint32 i=0; i<mLength; i++)
		if(subgraph[i]->GetLength() < mLength-1) {
			allConnected = false;
			break;
		}
	while(!allConnected && stats.diameter < 500) { // something must be wrong if we reach that 500..
		++stats.diameter;

		// the following is probably everything-but-smart-and-efficient, and should be re-written when speed becomes an issue.. (should use set operations)
		for(uint32 i=0; i<mLength; i++)
			for(uint32 j=i+1; j<mLength; j++)
				for(uint32 k=j+1; k<mLength; k++) {
					if(subgraph[i]->IsItemInSet(mSet[j]) && subgraph[i]->IsItemInSet(mSet[k]) && !subgraph[j]->IsItemInSet(mSet[k])) {
						subgraph[j]->AddItemToSet(mSet[k]);
						subgraph[k]->AddItemToSet(mSet[j]);
					}
					if(!subgraph[i]->IsItemInSet(mSet[j]) && subgraph[i]->IsItemInSet(mSet[k]) && subgraph[j]->IsItemInSet(mSet[k])) {
						subgraph[i]->AddItemToSet(mSet[j]);
						subgraph[j]->AddItemToSet(mSet[i]);
					}
					if(subgraph[i]->IsItemInSet(mSet[j]) && !subgraph[i]->IsItemInSet(mSet[k]) && subgraph[j]->IsItemInSet(mSet[k])) {
						subgraph[i]->AddItemToSet(mSet[k]);
						subgraph[k]->AddItemToSet(mSet[i]);
					}
				}

		allConnected = true;
		for(uint32 i=0; i<mLength; i++)
			if(subgraph[i]->GetLength() < mLength-1) {
				allConnected = false;
				break;
			}
	}

	// Compute average degree (in entire network)
	stats.avgDegree = 0.0;
	for(uint32 i=0; i<mLength; i++)
		stats.avgDegree += db->GetRow(mSet[i])->GetLength();
	stats.avgDegree /= mLength;

	// Compute edge density
	stats.edgeDensity = (float)mNumEdges / (mLength * (mLength-1) / 2);

	// Cleanup
	for(uint32 i=0; i<mLength; i++)
		delete subgraph[i];
	delete[] subgraph;

	return stats;
}
