#include "../global.h"

#include <Bass.h>
#include <Config.h>
#include <logger/Log.h>

#include <db/Database.h>
#include <itemstructs/ItemSet.h>

#include "../maxent/MaxEntModel.h"

#include "MaxEntInterestingness.h"

MaxEntInterestingness::MaxEntInterestingness(Config *config, Database *db) : Interestingness(config, db) {
	// Compute MaxEnt model
	LOG_MSG("MaxEnt - start building.");
	mModel = new MaxEntModel(config, db);
	LOG_MSG("MaxEnt - finished building.");
}
MaxEntInterestingness::~MaxEntInterestingness() {
	delete mModel;
}
void MaxEntInterestingness::LogParameters() {
	Interestingness::LogParameters();
}
void MaxEntInterestingness::ForPattern(ItemSet *pattern) {
	uint32 numNodes = pattern->GetLength();
	if(numNodes<=1) {
		pattern->SetNumEdges(0);
		pattern->SetLambda(0.0);
		return;
	}

	// Compute actual number of edges
	uint32 *nodes = pattern->GetSet();
	uint32 numEdges = 0;
	for(uint32 n=0; n<numNodes; n++)
		numEdges += mDatabase->GetRow(nodes[n])->IntersectionSize(pattern);
	numEdges /= 2;

	// Compute expected edge probability
	double edgeProb = 0.0; 
	for(uint32 i=0; i<numNodes; i++)
		for(uint32 j=i+1; j<numNodes; j++)
			edgeProb += mModel->probability(nodes[i], nodes[j]);
	edgeProb /= mMaxNumEdges[numNodes];

	// Set computed values
	pattern->SetNumEdges(numEdges);
	pattern->SetEdgeProbability(edgeProb);
	pattern->SetLambda((mMaxNumEdges[numNodes] * KL((double)numEdges / mMaxNumEdges[numNodes], edgeProb)) / 
		(numNodes * mDescIncludeCost + mDescExcludeCost));
}
void MaxEntInterestingness::ForExtension(ItemSet *pattern, ItemSet *parent, const uint32 node, const uint32 newNumNodes) {
	if(newNumNodes==1) {
		pattern->SetNumEdges(0);
		pattern->SetLambda(0.0);
		return;
	}
	
	// Compute expected edge probability starting from that of parent
	uint32 oldNumNodes = newNumNodes - 1;
	double edgeProb = parent->GetEdgeProbability() * mMaxNumEdges[oldNumNodes];
	uint32 *nodes = parent->GetSet();
	for(uint32 i=0; i<oldNumNodes; i++)
		edgeProb += mModel->probability(nodes[i], node);
	edgeProb /= mMaxNumEdges[newNumNodes];

	// Set values for new pattern
	uint32 numEdges = parent->GetNumEdges() +  mDatabase->GetRow(node)->IntersectionSize(parent);
	pattern->SetNumEdges(numEdges);
	pattern->SetEdgeProbability(edgeProb);
	pattern->SetLambda((mMaxNumEdges[newNumNodes] * KL((double)numEdges / mMaxNumEdges[newNumNodes], edgeProb)) / 
										(newNumNodes * mDescIncludeCost + mDescExcludeCost));
}
void MaxEntInterestingness::ForReduction(ItemSet *pattern, ItemSet *parent, const uint32 node, const uint32 newNumNodes) {
	if(newNumNodes==1) {
		pattern->SetNumEdges(0);
		pattern->SetLambda(0.0);
		return;
	}

	// Compute expected edge probability starting from that of parent
	uint32 oldNumNodes = newNumNodes + 1;
	double edgeProb = parent->GetEdgeProbability() * mMaxNumEdges[oldNumNodes];
	uint32 *nodes = parent->GetSet();
	for(uint32 i=0; i<oldNumNodes; i++)
		edgeProb -= mModel->probability(nodes[i], node);
	edgeProb /= mMaxNumEdges[newNumNodes];

	// Set values for new pattern
	uint32 numEdges = parent->GetNumEdges() - mDatabase->GetRow(node)->IntersectionSize(parent);
	pattern->SetNumEdges(numEdges);
	pattern->SetEdgeProbability(edgeProb);
	pattern->SetLambda((mMaxNumEdges[newNumNodes] * KL((double)numEdges / mMaxNumEdges[newNumNodes], edgeProb)) / 
		(newNumNodes * mDescIncludeCost + mDescExcludeCost));
}

// Update interestingness by adding given pattern to background knowledge, for iterative mining (ie update background distribution)
void MaxEntInterestingness::AddToBackground(ItemSet *pattern) {
	uint32 *items = pattern->GetValues();
	mModel->update(items, pattern->GetLength(), pattern->GetNumEdges());
	delete items;
}
