#include "../global.h"

#include <Bass.h>
#include <Config.h>
#include <logger/Log.h>

#include <db/Database.h>
#include <itemstructs/ItemSet.h>

#include "UniformInterestingness.h"

UniformInterestingness::UniformInterestingness(Config *config, Database *db) : Interestingness(config, db) {
	// Configuration
	mCacheUpTo = config->Read<uint32>("interestCache", 0); // default to no cache

	// Compute uniform expected edge probability
	mNumNodesUsed	= mNumNodes;
	mAlpha			= (double)db->GetNumItems() / mMaxNumEdges[mNumNodesUsed];

	// Create data structures for upper bound
	mNumDegreeCounts = mNumNodes + 2;
	mUBDegreeCounts = new uint32[mNumDegreeCounts];
	mUBDegreeCountsZeroes = new uint32[mNumDegreeCounts];
	for(uint32 i=0; i<mNumDegreeCounts; i++) 
		mUBDegreeCounts[i] = mUBDegreeCountsZeroes[i] = 0;

	// Create interestingness cache
	mCache = NULL;
	if(mCacheUpTo > 0) {
		mCacheUpTo = min(mCacheUpTo, db->GetNumRows()+1); // never cache more than needed, use [1,numRows]
		mCacheSize = mCacheUpTo;
		mCache = new double *[mCacheUpTo];
		for(uint32 numNodes=1; numNodes<mCacheUpTo; numNodes++) {
			mCache[numNodes] = new double[mMaxNumEdges[numNodes]+1];
			for(uint32 numEdges=0; numEdges<=mMaxNumEdges[numNodes]; numEdges++)
				mCache[numNodes][numEdges] = Compute(numEdges, numNodes);
		}
	}
}
UniformInterestingness::~UniformInterestingness() {
	delete[] mUBDegreeCounts;
	delete[] mUBDegreeCountsZeroes;
	for(uint32 i=1; i<mCacheUpTo; i++)
		delete[] mCache[i];
	delete[] mCache;
}
void UniformInterestingness::LogParameters() {
	Interestingness::LogParameters();
	char temp[20]; sprintf_s(temp, 20, "%f", mAlpha);
	LOG_MSG("Expected density:\t" + string(temp));
	sprintf_s(temp, 20, "%d", mCacheUpTo);
	LOG_MSG("Cache up to:\t\t" + string(temp) + " nodes");
}
void UniformInterestingness::ForPattern(ItemSet *pattern) {
	uint32 numNodes = pattern->GetLength();
	if(numNodes==0) {
		pattern->SetNumEdges(0);
		pattern->SetLambda(0.0);
		return;
	}

	uint32 *nodes = pattern->GetSet();
	uint32 numEdges = 0;
	for(uint32 n=0; n<numNodes; n++)
		numEdges += mDatabase->GetRow(nodes[n])->IntersectionSize(pattern);
	numEdges /= 2;

	pattern->SetNumEdges(numEdges);
	pattern->SetLambda(GetItFor(numEdges, numNodes));
}
void UniformInterestingness::ForExtension(ItemSet *pattern, ItemSet *parent, const uint32 node, const uint32 newNumNodes) {
	uint32 numEdges = parent->GetNumEdges() + mDatabase->GetRow(node)->IntersectionSize(parent);
	pattern->SetNumEdges(numEdges);
	pattern->SetLambda(GetItFor(numEdges, newNumNodes));
}
void UniformInterestingness::ForReduction(ItemSet *pattern, ItemSet *parent, const uint32 node, const uint32 newNumNodes) {
	uint32 numEdges = parent->GetNumEdges() - mDatabase->GetRow(node)->IntersectionSize(parent);
	pattern->SetNumEdges(numEdges);
	pattern->SetLambda(GetItFor(numEdges, newNumNodes));
}
void UniformInterestingness::AddToBackground(ItemSet *pattern) {
	// No sophisticated background model -- simply remove pattern from graph completely
	mDatabase->RemoveNodes(pattern);

	// Update number of nodes in use
	mNumNodesUsed -= pattern->GetLength();

	// Update uniform expected edge probability
	mAlpha			= (double)mDatabase->GetNumItems() / mMaxNumEdges[mNumNodesUsed];

	// Update interestingness cache
	if(mCacheUpTo > 0) {
		mCacheSize = min(mCacheUpTo, mNumNodesUsed+1);
		for(uint32 numNodes=1; numNodes<mCacheSize; numNodes++)
			for(uint32 numEdges=0; numEdges<=mMaxNumEdges[numNodes]; numEdges++)
				mCache[numNodes][numEdges] = Compute(numEdges, numNodes);
	}
}

void UniformInterestingness::ComputeUpperBound(ItemSet *pattern, const uint32 fromNode) {
	// Reset counts to all zeroes
	memcpy_s(mUBDegreeCounts, mNumDegreeCounts*sizeof(uint32), mUBDegreeCountsZeroes, mNumDegreeCounts*sizeof(uint32));

	// Count # degrees
	uint32 maxDegreeToW = 0;
	uint32 maxDegreeInX = 0;
	for(uint32 n=fromNode; n<mNumNodes; n++) {
		ItemSet *row = mDatabase->GetRow(n);
		if(row->GetLength() == 0)
			continue; // skip if empty

		// To W
		uint32 d = row->IntersectionSize(pattern);
		++mUBDegreeCounts[d];
		maxDegreeToW = max(maxDegreeToW, d);

		// In X
		d = row->CountItemsGEQThan(fromNode);
		++mUBDegreeCounts[mNumDegreeCounts - d - 1];
		maxDegreeInX = max(maxDegreeInX, d);
	}

	// Loop over candidate sizes increasingly
	// Compute bound for each size
	double ub = 0.0;
	uint32 numCands = mNumNodes - fromNode;
	uint32 numEdges = pattern->GetNumEdges(), numEdgesInX = 0;
	uint32 numNodes = pattern->GetLength();
	for(uint32 newNodes=1; newNodes<=numCands; newNodes++) {
		// Update numEdges and numNodes
		if(maxDegreeToW > 0) {
			numEdges += maxDegreeToW;
			--mUBDegreeCounts[maxDegreeToW];
			while(mUBDegreeCounts[maxDegreeToW] == 0 && maxDegreeToW > 0)
				--maxDegreeToW;
		}
		if(maxDegreeInX > 0) {
			numEdgesInX += maxDegreeInX;
			--mUBDegreeCounts[mNumDegreeCounts - maxDegreeInX - 1];
			while(mUBDegreeCounts[mNumDegreeCounts - maxDegreeInX - 1] == 0 && maxDegreeInX > 0)
				--maxDegreeInX;
		}
		++numNodes;

		// Compute bound and decide what to do next
		uint32 e = numEdges + numEdgesInX / 2;
		double b = e > mMaxNumEdges[numNodes] ? GetItFor(mMaxNumEdges[numNodes], numNodes) : GetItFor(e, numNodes);
		if(b > ub)
			ub = b;
		if(maxDegreeToW == 0 && maxDegreeInX == 0)
			break; // things can only get worse..
	}

	// We're done!
	pattern->SetUpperBound(ub);
}


/*
-- Correct but currently not used by any of the search implementations.
bool UniformInterestingness::CanSingleExtensionImprove(ItemSet *pattern, const uint32 fromNode, const uint32 toNode, const uint32 newNumNodes) {
	uint32 numEdges = pattern->GetNumEdges();

	// For all possible extensions indicated
	uint32 maxNewEdgesSeen = 0;
	for(uint32 node=fromNode; node<toNode; node++) {
		// Check whether already there
		if(pattern->IsItemInSet(node))
			continue;

		// Compute would-be lambda
		uint32 newEdges = mDatabase->GetRow(node)->IntersectionSize(pattern);
		if(newEdges	> maxNewEdgesSeen) { // only need to test if adds more edges
			if(GetItFor(numEdges + newEdges, newNumNodes) > pattern->GetLambda())
				return true;
			maxNewEdgesSeen = newEdges;
		}
	}
	return false;
}*/

