#include "../global.h"

#include <Bass.h>
#include <Config.h>
#include <logger/Log.h>

#include <db/Database.h>
#include <algorithm>
#include <iterator>

#include "MaxEntModel.h"

using namespace std;

MaxEntModel::MaxEntModel(Config *config, Database *db) {
	mConfig	  = config;
	mDatabase = db;
	mNumNodes = db->GetNumRows();
	vector<int> marginals;
	marginals.reserve(mNumNodes);

	for (uint32 i=0; i<mNumNodes; i++){
		marginals.push_back(db->GetRow(i)->GetLength());
		//marginals.push_back(sizeof(db->GetRow(i))/sizeof(*(db->GetRow(i))));
	}

	vector<int> marginalsu;
	vector<vector<int> > marginIndices;
	vector<int> nMult;

	makeUnique(&marginals, marginalsu, marginIndices);

	int nM = marginalsu.size();

	for (int i=0; i<nM; i++){
		nMult.push_back(marginIndices[i].size());
	}

	vector<double> lambdaq;
	lambdaq.reserve(nM);

	for (int i=0; i<nM; i++){
		lambdaq.push_back(0);
	}
    
	set<const uint32> temp;
	for (unsigned i=0; i<mNumNodes; i++){
        mNodeToSolution.push_back(temp);
    }
    
	// Start of the iteration
	int counter = 0;
	double factors[10] = {0.1, 0.2, 0.3, 0.4, 0.55, 0.7, 0.85, 1, 1.2, 1.5};
	double infeasibility=1;

	while (counter < 100 && infeasibility>mNumNodes*1e-12){
		//cout << "Iteration: " << counter+1 << endl;
		// Compute gradient and Hessian diagonal
		vector<double> gLambdaq;
		gLambdaq.resize(nM); // was .reserve(nM)
		vector<double> hLambdaq;
		hLambdaq.resize(nM); // was .reserve(nM)

		//initialise
		for (int i=0; i<nM; i++) {
			gLambdaq[i]=-marginalsu[i];
			hLambdaq[i]=0;
		}

		for (int i=0; i<nM; i++){
			for (int j=0; j<nM; j++){
				double v=exp(lambdaq[i]+lambdaq[j]);
				double g = v/(1+v);
				double h = v/((1+v)*(1+v));
				if (i==j){
					gLambdaq[i]+=g*(nMult[j]-1);
					hLambdaq[i]+=h*(nMult[j]-1);
				}
				else {
					gLambdaq[i]+=g*nMult[j];
					hLambdaq[i]+=h*nMult[j];
				}
			}
		}

		// Compute step direction
		vector<double> dLambdaq;
		dLambdaq.resize(nM); // was .reserve(nM)
		for (int j=0; j<nM; j++)
			dLambdaq[j]=-gLambdaq[j]/hLambdaq[j];

		//find the best step length

		double bestFactor = 0;
		double bestError=0;

		for (int j=0; j<nM; j++)
			bestError -= lambdaq[j]*marginalsu[j]*nMult[j];

		for (int i=0; i<nM; i++){
			for (int j=0; j<nM; j++){
				if (i==j)
					bestError +=0.5*log(1+exp(lambdaq[i]+lambdaq[j]))*nMult[i]*(nMult[j]-1);
				else
					bestError +=0.5*log(1+exp(lambdaq[i]+lambdaq[j]))*nMult[i]*nMult[j];
			}
		}

		for (int k=0; k<10; k++)
		{
			double factor = factors[k];
			vector<double> lambdaqTry;
			lambdaqTry.resize(nM); // was .reserve(nM)

			for (int j=0; j<nM; j++)
				lambdaqTry[j]=lambdaq[j]+dLambdaq[j]*factor;

			double errorTry = 0;

			for (int j=0; j<nM; j++)
				errorTry -= lambdaqTry[j]*marginalsu[j]*nMult[j];

			for (int i=0; i<nM; i++){
				for (int j=0; j<nM; j++){
					if (i==j)
						errorTry += 0.5*log(1+exp(lambdaqTry[i]+lambdaqTry[j]))*nMult[i]*(nMult[j]-1);
					else
						errorTry += 0.5*log(1+exp(lambdaqTry[i]+lambdaqTry[j]))*nMult[i]*nMult[j];
				}
			}


			if (errorTry<bestError)
			{
				bestError=errorTry;
				bestFactor = factor;
			}
			else break;
		}

		//new lambdas
		for (int j=0; j<nM; j++)
		{
			lambdaq[j] += dLambdaq[j]*bestFactor;
		}
		counter++;

		//Compute the entropy -- the primal objective
		double entropy = 0;
		double dual=0;
		double lambdasum = 0;

		for (int i=0; i<nM; i++)
		{
			lambdasum+=lambdaq[i]*marginalsu[i]*nMult[i];
			for (int j=0; j<nM; j++)
			{
				double p = exp(lambdaq[i]+lambdaq[j])/(1+exp(lambdaq[i]+lambdaq[j]));
				if (i==j){
					entropy = entropy - 0.5*(p*log(p) + (1-p)*log(1-p))*nMult[i]*(nMult[j]-1);
					dual+=0.5*log(1+exp(lambdaq[i]+lambdaq[j]))*nMult[i]*(nMult[j]-1);
				}
				else{
					entropy = entropy - 0.5*(p*log(p) + (1-p)*log(1-p))*nMult[i]*nMult[j];
					dual+=0.5*log(1+exp(lambdaq[i]+lambdaq[j]))*nMult[i]*nMult[j];
				}
			}
		}
		dual=dual-lambdasum;
		//cout<<"Dual: "<<dual<<endl;
		//cout << "Best factor: " << bestFactor << " - Current cost: " << bestError << " - Entropy: " << entropy << endl << endl;
		// The following line is to decrease the smallest possible stepsize if the optimal one was 0.
		if (bestFactor == 0) factors[0] = factors[0]/2;

		infeasibility = 0;
		for (int j=0; j<nM; j++)
			infeasibility += gLambdaq[j]*gLambdaq[j]*nMult[j];
		//cout << "Gradient norm squared: " << infeasibility << endl;

	}

	//Debug lines
	/*for (int i=0; i<nM; i++){
		double difference=0;
		for (int j=0; j<nM; j++){
			if (i!=j)
				difference+= (exp(lambdaq[i]+lambdaq[j])/(1+exp(lambdaq[i]+lambdaq[j])))*nMult[j];
			else
				difference+= (exp(lambdaq[i]+lambdaq[j])/(1+exp(lambdaq[i]+lambdaq[j])))*(nMult[j]-1);
		}
		cout<<difference<<" "<<marginalsu[i]<<endl;
	}*/

	//vector<double> lambda;
    mLambda.resize(mNumNodes); // was .reserve()
    
	for (int j=0; j<nM; j++){
		for (unsigned int k=0; k<marginIndices[j].size(); k++){
			mLambda[marginIndices[j][k]]=lambdaq[j];
		}
	}

	marginIndices.clear();
    sort(lambdaq.begin(),lambdaq.end());
	mLargestLambda=lambdaq[lambdaq.size()-1];
	mSmallestLambda=lambdaq[0];
	lambdaq.clear();
    
    /*for (int i=0; i<mNumNodes; i++){
		for (int j=i+1; j<mNumNodes; j++){
            mExpLambda[i][j]=exp(lambda[i]+lambda[j]);
        }
    }
    lambda.clear();*/
}
MaxEntModel::~MaxEntModel() {

}

double MaxEntModel::probability(int nodea, int nodeb){
	if (nodea!=nodeb){
		//double n= nodeb>nodea ? mExpLambda[nodea][nodeb] : mExpLambda[nodeb][nodea];
		//return n/(1+n);
        set<const uint32> solutions;
        set<const uint32>::const_iterator it;
        double n;
        insert_iterator<set<const uint32> > ins_it(solutions, solutions.begin());
        set_intersection(mNodeToSolution[nodea].begin(), mNodeToSolution[nodea].end(), mNodeToSolution[nodeb].begin(), mNodeToSolution[nodeb].end(), ins_it);
        if (!solutions.empty()){
            n=1;
            for (it=solutions.begin(); it!=solutions.end(); it++){
                n=n*exp(mLambdaS[*it]);
            }
            n=n*exp(mLambda[nodea])*exp(mLambda[nodeb]);
            
        } else{
            n=exp(mLambda[nodea])*exp(mLambda[nodeb]);
        }
        solutions.clear();
		return n/(1+n);
	}
	return 0;
}

void MaxEntModel::update(const uint32 *nodes, const uint32 numNodes, const uint32 numEdges){
	double a;
	double b;

	if (fabs(mSmallestLambda)>fabs(mLargestLambda)){
		a = -3*fabs(mSmallestLambda);
		b = 3*fabs(mSmallestLambda);
	}
	else {
		a = -3*fabs(mLargestLambda);
		b = 3*fabs(mLargestLambda);
	}
    
    double f_a;
    double f_b;
    double f_c;
    double c;
    double v_a;
    double v_b;
    double v_c;
    
    
    map<const uint32,map<const uint32,double> > expLambda;
    for (unsigned int i=0; i<numNodes; i++){
        for (unsigned int j=i+1; j<numNodes; j++){
            set<const uint32> solutions;
            set<const uint32>::const_iterator it;
            insert_iterator<set<const uint32> > ins_it(solutions, solutions.begin());
            set_intersection(mNodeToSolution[nodes[i]].begin(), mNodeToSolution[nodes[i]].end(), mNodeToSolution[nodes[j]].begin(), mNodeToSolution[nodes[j]].end(), ins_it);
            if (!solutions.empty()){
                double r=1;
                for (it=solutions.begin(); it!=solutions.end(); it++){
                    r=r*exp(mLambdaS[*it]);
                }
                expLambda[i][j]=exp(mLambda[nodes[i]])*exp(mLambda[nodes[j]])*r;
            }
            else{
                expLambda[i][j]=exp(mLambda[nodes[i]])*exp(mLambda[nodes[j]]);
            }
            solutions.clear();
        }
    }
    
    for (unsigned int i=0; i<numNodes; i++){
        mNodeToSolution[nodes[i]].insert(mLambdaS.size());
    }
    
    while (b-a>10e-12) {
        f_a=0;
        f_b=0;
        f_c=0;
        c=(a+b)/2;
        
        for (unsigned int i=0; i<numNodes; i++){
            for (unsigned int j=i+1; j<numNodes; j++){
                v_a=expLambda[i][j]*exp(a);
                v_b=expLambda[i][j]*exp(b);
                v_c=expLambda[i][j]*exp(c);
                f_a+=v_a/(1+v_a);
                f_b+=v_b/(1+v_b);
                f_c+=v_c/(1+v_c);
            }
        }
        f_a=f_a-numEdges;
        f_b=f_b-numEdges;
        f_c=f_c-numEdges;
        
        if (f_c<0)
            a=c;
        else
            b=c;
    }
    
    double lambdac=(a+b)/2;
    mLambdaS.push_back(lambdac);
    expLambda.clear();
    
}
void MaxEntModel::makeUnique(vector<int>* sums, vector<int>& uniqueSums, vector<vector<int> >& indices){
	uniqueSums.clear();
	for (unsigned int i=0; i<indices.size(); i++) indices.at(i).clear();
	indices.clear();

	for (unsigned int i=0; i<sums->size(); i++)
	{
		int x=sums->at(i);
		unsigned int k;
		for (k=0 ; k<uniqueSums.size(); k++)
		{
			if (x==uniqueSums.at(k))
			{
				indices.at(k).push_back(i);
				break;
			}
		}
		if (k==uniqueSums.size())
		{
			uniqueSums.push_back(x);
			vector<int> temp;
			temp.push_back(i);
			indices.push_back(temp);
			temp.clear();
		}
	}
}

