#include "../global.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Config.h>
#include <TimeUtils.h>
#include <logger/Log.h>

#include <db/Database.h>
#include <collections/PatternCollection.h>

#include "../interest/Interestingness.h"

// #include "best/BestFirstSearch.h"
#include "best/BetterBestFirstSearch.h" // supersedes BestFirstSearch --> better performance.
#include "best/BoundedBestFirstSearch.h"
#include "hillclimber/Hillclimber.h"

#include "Search.h"

Search::Search(Config *config, Database *db, PatternCollection *patternCollection) {
	mConfig = config;
	mDatabase = db->Clone(); // create copy --> search may change the data (in particular when doing iterative mining)
	mPatternCollection = patternCollection;
	mInterestingness = Interestingness::Create(config, mDatabase);

	mMinInterestingness = config->Read<double>("minInterest", 0.0);
	mMinNumNodes		= config->Read<uint32>("minNumNodes", 1);
	mMaxSize			= config->Read<uint32>("maxSize", UINT32_MAX_VALUE);
	mDisableUpperBound	= config->Read<bool>("disableUB", false);
	mUseUpperBound		= !mDisableUpperBound && mInterestingness->HasUpperBound();

	mMaxSize = min(mMaxSize, mDatabase->GetNumRows());
	mCandidateCount = 0;
	mMaxTime = config->Read<uint32>("maxTime", 0) * 60; // from minutes to seconds
	mTimer = NULL;
}

Search::~Search() {
	delete mDatabase;
	delete mTimer;
	delete mInterestingness;
}
void Search::LogParameters() {
	mInterestingness->LogParameters();
	char temp[20]; sprintf_s(temp, 20, "%f", mMinInterestingness);
	LOG_MSG("Min. interestingness:\t" + string(temp));
	//LOG_MSG("Max time:\t\t" + mConfig->Read<string>("maxTime", 0));
	LOG_MSG("Search type:\t\t" + mConfig->Read<string>("searchType"));
	LOG_MSG("Mine:\t\t\t" + mConfig->Read<string>("mine"));
	//LOG_MSG("Min num nodes:\t\t" + mConfig->Read<string>("minNumNodes", 1));
	LOG_MSG("Max depth:\t\t" + mConfig->Read<string>("maxDepth", string("none")));
	LOG_MSG(string("Using upper bound:\t") + (CanUseUpperBound() && mUseUpperBound ? "yes" : "no"));
}
Search* Search::Create(Config *config, Database *db, PatternCollection *pCollection) {
	string desc = config->Read<string>("searchType");
	return Create(desc, config, db, pCollection);
}
Search* Search::Create(const string &desc, Config *config, Database *db, PatternCollection *pCollection) {
	if(desc.compare("best") == 0)
		return new BetterBestFirstSearch(config, db, pCollection);
	if (desc.compare("bounded-best") == 0) // bounded best first search
		return new BoundedBestFirstSearch(config, db, pCollection);
	if(desc.compare("hillclimber") == 0)
		return new Hillclimber(config, db, pCollection);
	THROW("Unknown search type.");
}
uint32 Search::GetElapsedTime() {
	return (uint32)(mTimer->GetElapsedTime()); // sec
}
void Search::InitTimer() {
	delete mTimer;
	mTimer = new Timer();	
}
bool Search::OutOfTime() {
	if(mMaxTime == 0)
		return false;
	return mTimer->GetElapsedTime() > mMaxTime;
}
ItemSet *Search::ExtendPattern(ItemSet *pattern, const uint32 addNode, const uint32 newNumNodes) {
	// Create extension
	ItemSet *extension = pattern->Clone();
	extension->AddItemToSet(addNode);

	// Compute quality
	mInterestingness->ForExtension(extension, pattern, addNode, newNumNodes);

	return extension;
}
ItemSet *Search::ShrinkPattern(ItemSet *pattern, const uint32 removeNode, const uint32 newNumNodes) {
	// Create extension
	ItemSet *reduction = pattern->Clone();
	reduction->RemoveItemFromSet(removeNode);

	// Compute quality
	mInterestingness->ForReduction(reduction, pattern, removeNode, newNumNodes);

	return reduction;
}
