#include "../../global.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Config.h>
#include <logger/Log.h>

#include <db/Database.h>

#include "../interest/Interestingness.h"

#include "BestFirstSearch.h"

BestFirstSearch::BestFirstSearch(Config *config, Database *db, PatternCollection *pColl) : Search(config, db, pColl) {
	// Configuration

}
BestFirstSearch::~BestFirstSearch() {

}
void BestFirstSearch::LogParameters() {
	Search::LogParameters();
}
uint64 BestFirstSearch::DoSearch() {
	InitTimer();

	// Init some structures
	PatternQueue *pQueue = new PatternQueue();
	uint32 numNodes = mDatabase->GetNumRows();
	double minLambda = mPatternCollection->GetMinThreshold();

	// Init beam with empty root candidate
	{
		ItemSet *root = ItemSet::CreateEmpty(Uint32ItemSetType, numNodes);
		pQueue->push(root);
	}

	// While candidates and within the allowed time
	while(!pQueue->empty() && !OutOfTime()) {
		// Get next pattern and remove from queue
		ItemSet *pattern = pQueue->top(); pQueue->pop();

		// Check if we need to proceed any further
		if(mUseUpperBound && pattern->GetUpperBound() < minLambda) {
			// All candidates after this will have lower ub, so abort
			delete pattern;
			break;
		}

		// Construct all possible extensions
		uint32 newPatLen = pattern->GetLength() + 1;
		uint32 fromNode = newPatLen==1 ? 0 : pattern->GetLastItem() + 1;
		for(uint32 c=fromNode; c<numNodes; c++) {
			// Check if candidate node is connected to any other node (ie is it still present?)
			if(mDatabase->GetRow(c)->GetLength() == 0)
				continue;

			// Construct candidate pattern
			ItemSet *cand = ExtendPattern(pattern, c, newPatLen);

			// Increase counter
			++mCandidateCount;

			// Output pattern and update minLambda if a local optimum ---> This only helps if CanSingleExtensionImprove() can be easily computed.
			//																Since this is not the case with MaxEnt and we are now mostly interested in top-1 search, this is disabled
			/*
			if(newPatLen==mMaxDepth || !mInterestingness->CanSingleExtensionImprove(cand, 0, fromNode, newPatLen+1)) {
				OutputPattern(cand);
				minLambda = max(minLambda, mPatternCollection->GetMinThreshold());
			}
			*/
			
			// Add pattern to PatternCollection (which in practice means: maintain top-k, where k is often 1)
			mPatternCollection->Add(cand);
			// and update minimum to be used in conjunction with upperbound
			minLambda = mPatternCollection->GetMinThreshold();

			if(newPatLen < mMaxSize) { // Do we need to consider further extensions?
				// Check whether pruning is enabled
				if(mUseUpperBound) {
					// Do upper bound stuff
					mInterestingness->ComputeUpperBound(cand, c+1);

					// Check whether viable extensions exist
					if(cand->GetUpperBound() > minLambda) {
						// Insert into priority queue
						pQueue->push(cand);
					} else {
						// We can delete it immediately
						delete cand;
					}
				} else {
					// No pruning - always insert into priority queue
					pQueue->push(cand);
				}
			} else { // maxDepth reached
				delete cand;
			}
		}

		// Delete current pattern
		delete pattern;

		// Cleanup at the end of the queue
		// Impossible with current pQueue...
	}

	// Report to the user when needed
	if(OutOfTime())
		LOG_MSG("** Out of time! Stopping right here, right now. **");

	// Cleanup
	while(!pQueue->empty()) {
		ItemSet *pattern = pQueue->top(); 
		pQueue->pop();
		delete pattern;
	}
	delete pQueue;

	// Return resulting subgroups
	return mCandidateCount;
}
