#include "../../global.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Config.h>
#include <logger/Log.h>

#include <db/Database.h>

#include "../interest/Interestingness.h"

#include "BetterBestFirstSearch.h"

BetterBestFirstSearch::BetterBestFirstSearch(Config *config, Database *db, PatternCollection *pColl) : Search(config, db, pColl) {
	// Init
	mQueue = new BetterPatternQueue();
}
BetterBestFirstSearch::~BetterBestFirstSearch() {
	delete mQueue;
}
void BetterBestFirstSearch::LogParameters() {
	Search::LogParameters();
}
uint64 BetterBestFirstSearch::DoSearch() {
	InitTimer();

	// Init some structures
	uint32 numNodes = mDatabase->GetNumRows();
	double minLambda = mPatternCollection->GetMinThreshold();

	// Init beam with empty root candidate
	Insert(ItemSet::CreateEmpty(Uint32ItemSetType, numNodes));

	// While candidates and within the allowed time
	while(!mQueue->empty() && !OutOfTime()) {
		// Get next pattern and remove from queue
		ItemSet *pattern = GetNext();

		// Check if we need to proceed any further
		if(mUseUpperBound && pattern->GetUpperBound() < minLambda) {
			// All candidates after this will have lower ub, so abort
			delete pattern;
			break;
		}

		// Construct all possible extensions
		uint32 newPatLen = pattern->GetLength() + 1;
		uint32 fromNode = newPatLen==1 ? 0 : pattern->GetLastItem() + 1;
		for(uint32 c=fromNode; c<numNodes; c++) {
			// Check if candidate node is connected to any other node (ie is it still present?)
			if(mDatabase->GetRow(c)->GetLength() == 0)
				continue;

			// Construct candidate pattern
			ItemSet *cand = ExtendPattern(pattern, c, newPatLen);

			// Increase counter
			++mCandidateCount;

			// Add pattern to PatternCollection (which in practice means: maintain top-k, where k is often 1)
			mPatternCollection->Add(cand);
			// and update minimum to be used in conjunction with upperbound
			if (mPatternCollection->GetMinThreshold() > minLambda) {
				minLambda = mPatternCollection->GetMinThreshold();
				PruneQueue(minLambda); // make sure the queue is no longer than needed
			}

			if(newPatLen < mMaxSize) { // Do we need to consider further extensions?
				// Check whether pruning is enabled
				if(mUseUpperBound) {
					// Do upper bound stuff
					mInterestingness->ComputeUpperBound(cand, c+1);

					// Check whether viable extensions exist
					if(cand->GetUpperBound() > minLambda) {
						// Insert into priority queue
						Insert(cand);
					} else {
						// We can delete it immediately
						delete cand;
					}
				} else {
					// No pruning - always insert into priority queue
					Insert(cand);
				}
			} else { // maxDepth reached
				delete cand;
			}
		}

		// Delete current pattern
		delete pattern;
	}

	// Report to the user when needed
	if(OutOfTime())
		LOG_MSG("** Out of time! Stopping right here, right now. **");

	// Cleanup
	while(!mQueue->empty())
		delete GetNext(); 

	// Return resulting subgroups
	return mCandidateCount;
}

void BetterBestFirstSearch::Insert(ItemSet *pattern) {
	mQueue->insert(pattern);
}
ItemSet *BetterBestFirstSearch::GetNext() {
	BetterPatternQueue::iterator it = mQueue->begin();
	ItemSet *is = *it;
	mQueue->erase(it);
	return is;
}
void BetterBestFirstSearch::PruneQueue(double minInterestingness) {
	if (mQueue->size() < 2)
		return;

	BetterPatternQueue::iterator it = mQueue->end();
	--it;
	while (it!=mQueue->begin() && (*it)->GetUpperBound() < minInterestingness) { // doesn't delete the last one, even if needed, but we don't worry about that
		BetterPatternQueue::iterator tmp = it;
		--it;
		delete *tmp;
		mQueue->erase(tmp);
	}

	//for (BetterPatternQueue::iterator it = mQueue->begin(); it != mQueue->end(); it++)
	//	printf("%.03f ", (*it)->GetUpperBound());
	//printf("\n\n\n");
}
