#include "../../global.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Config.h>
#include <logger/Log.h>

#include <db/Database.h>

#include "../interest/Interestingness.h"

#include "BoundedBestFirstSearch.h"

BoundedBestFirstSearch::BoundedBestFirstSearch(Config *config, Database *db, PatternCollection *pColl) : Search(config, db, pColl) {
	// Configuration
	mMaxQueueSize = config->Read<uint32>("bestMaxQueue");
	if(mMaxQueueSize == 0)
		THROW("Maximum queue size should be more than 0.");

	// Init
	mQueue = new BoundedPatternQueue();
	mQueueSize = 0;
}
BoundedBestFirstSearch::~BoundedBestFirstSearch() {
	delete mQueue;
}
void BoundedBestFirstSearch::LogParameters() {
	Search::LogParameters();
	char temp[100]; sprintf_s(temp, 100, "%u", mMaxQueueSize);
	LOG_MSG("Max queue size:\t\t" + string(temp));
}
uint64 BoundedBestFirstSearch::DoSearch() {
	InitTimer();

	// Init some structures
	uint32 numNodes = mDatabase->GetNumRows();
	double minLambda = mPatternCollection->GetMinThreshold();

	// Init beam with empty root candidate
	Insert(ItemSet::CreateEmpty(Uint32ItemSetType, numNodes));

	// While candidates and within the allowed time
	while(mQueueSize > 0 && !OutOfTime()) {
		// Get next pattern and remove from queue
		ItemSet *pattern = GetNext();

		// Check if we need to proceed any further
		if(mUseUpperBound && pattern->GetUpperBound() < minLambda) {
			// All candidates after this will have lower ub, so abort
			delete pattern;
			break;
		}

		// Construct all possible extensions
		uint32 newPatLen = pattern->GetLength() + 1;
		uint32 fromNode = newPatLen==1 ? 0 : pattern->GetLastItem() + 1;
		for(uint32 c=fromNode; c<numNodes; c++) {
			// Check if candidate node is connected to any other node (ie is it still present?)
			if(mDatabase->GetRow(c)->GetLength() == 0)
				continue;

			// Construct candidate pattern
			ItemSet *cand = ExtendPattern(pattern, c, newPatLen);

			// Increase counter
			++mCandidateCount;

			// Add pattern to PatternCollection (which in practice means: maintain top-k, where k is often 1)
			mPatternCollection->Add(cand);
			// and update minimum to be used in conjunction with upperbound
			minLambda = mPatternCollection->GetMinThreshold();

			if(newPatLen < mMaxSize) { // Do we need to consider further extensions?
				// Check whether pruning is enabled
				if(mUseUpperBound) {
					// Do upper bound stuff
					mInterestingness->ComputeUpperBound(cand, c+1);

					// Check whether viable extensions exist
					if(cand->GetUpperBound() > minLambda) {
						// Insert into priority queue
						Insert(cand);
					} else {
						// We can delete it immediately
						delete cand;
					}
				} else {
					// No pruning - always insert into priority queue
					Insert(cand);
				}
			} else { // maxDepth reached
				delete cand;
			}
		}

		// Delete current pattern
		delete pattern;
	}

	// Report to the user when needed
	if(OutOfTime())
		LOG_MSG("** Out of time! Stopping right here, right now. **");

	// Cleanup
	while(mQueueSize > 0)
		delete GetNext(); 

	// Return resulting subgroups
	return mCandidateCount;
}

void BoundedBestFirstSearch::Insert(ItemSet *pattern) {
	// If the max queue size has been reached
	if(mQueueSize==mMaxQueueSize) {
		// Then check whether pattern should be inserted or not
		if(pattern->GetUpperBound() > mQueue->back()->GetUpperBound()) {
			// Yes, better, so delete current last
			ItemSet *is = mQueue->back();
			mQueue->pop_back();
			delete is;
		} else {
			// No, current pattern can be skipped; move on, nothing to see here
			delete pattern;
			return;
		}
	} else {
		mQueueSize++;
	}

	// Still here? Do actual insertion
	BoundedPatternQueue::iterator it = mQueue->begin();
	while(it!=mQueue->end() && (*it)->GetUpperBound() > pattern->GetUpperBound())
		++it;
	mQueue->insert(it, pattern);
}
ItemSet *BoundedBestFirstSearch::GetNext() {
	ItemSet *is = mQueue->front();
	--mQueueSize;
	mQueue->pop_front();
	return is;
}
