#include "../../global.h"

#if defined (__unix__)
#include <glibc_s.h>
#endif

#include <Config.h>
#include <RandomUtils.h>
#include <logger/Log.h>

#include <db/Database.h>
#include <collections/TopKCollection.h>

#include "../interest/Interestingness.h"

#include "Hillclimber.h"

Hillclimber::Hillclimber(Config *config, Database *db, PatternCollection *pColl) : Search(config, db, pColl) {
	// Configuration
	mNumRuns = config->Read<uint32>("climbNumRuns", 1);
	mSeedMode = SEED_UNIFORM;
	string sSeedMode = config->Read<string>("seedMode", "uniform");
	if(mNumRuns >= db->GetNumRows() || sSeedMode.compare("all") == 0) {
		mSeedMode = SEED_ALL;
		mNumRuns = db->GetNumRows();
	} else if(sSeedMode.compare("degree") == 0)
		mSeedMode = SEED_DEGREE;
	else if(sSeedMode.compare("interestingness") == 0)
		mSeedMode = SEED_INTERESTINGNESS;
}
Hillclimber::~Hillclimber() {

}
void Hillclimber::LogParameters() {
	Search::LogParameters();
	char temp[20]; sprintf_s(temp, 20, "%u", mNumRuns);
	LOG_MSG("# hillclimb runs:\t" + string(temp));
	string sSeedMode = "all";
	switch(mSeedMode) {
	case SEED_UNIFORM:
		sSeedMode = "uniform";
		break;
	case SEED_DEGREE:
		sSeedMode = "degree";
		break;
	case SEED_INTERESTINGNESS:
		sSeedMode = "interestingness";
		break;
	}
	LOG_MSG("Seed mode:\t\t" + sSeedMode);
}
uint64 Hillclimber::DoSearch() {
	InitTimer();

	// Init some structures
	uint32 numNodes = mDatabase->GetNumRows();
	ItemSet *connectedNodes = ItemSet::CreateEmpty(Uint32ItemSetType, numNodes);

	// Select seed nodes
	uint32 *seedNodes = new uint32[mNumRuns];
	switch(mSeedMode) {
		case SEED_ALL:
			if(mNumRuns!=numNodes)
				THROW("Number of runs should be equal to number of nodes here.");
			for(uint32 r=0; r<mNumRuns; r++)
				seedNodes[r] = r;
			break;
	
		case SEED_UNIFORM:
			{
				ItemSet *selected = ItemSet::CreateEmpty(Uint32ItemSetType, numNodes);
				uint32 n=0;
				while(n<mNumRuns) {
					uint32 node = RandomUtils::UniformUint32(numNodes);
					if(!selected->IsItemInSet(node)) {
						seedNodes[n++] = node;
						selected->AddItemToSet(node);
					}
				}
				delete selected;
			}
			break;

		case SEED_DEGREE:	// Cannot rely on item order in iterative mining situation (possibly removing nodes every iteration)
			{
				TopKCollection *candCollection = new TopKCollection(mNumRuns, mDatabase);
				for(uint32 i=0; i<numNodes; i++) {
					ItemSet *pattern = mDatabase->GetRow(i)->Clone();
					pattern->SetId(i);
					pattern->SetLambda(pattern->GetLength());
					candCollection->Add(pattern);
					delete pattern;
				}
				ItemSet **topCandidates = candCollection->GetPatterns();
				for(uint32 r=0; r<candCollection->GetNumPatterns(); r++)
					seedNodes[r] = topCandidates[r]->GetId();
				delete candCollection;
			}
			break;

		case SEED_INTERESTINGNESS:
			{
				TopKCollection *candCollection = new TopKCollection(mNumRuns, mDatabase);
				for(uint32 i=0; i<numNodes; i++) {
					ItemSet *pattern = mDatabase->GetRow(i)->Clone();
					pattern->SetId(i);
					mInterestingness->ForPattern(pattern);
					candCollection->Add(pattern);
					delete pattern;
				}
				ItemSet **topCandidates = candCollection->GetPatterns();
				for(uint32 r=0; r<candCollection->GetNumPatterns(); r++)
					seedNodes[r] = topCandidates[r]->GetId();
				delete candCollection;
			}
			break;
	}
	
	// Do requested runs, using seed nodes as just determined
	for(uint32 r=0; r<mNumRuns && !OutOfTime(); r++) {
		// Init candidate solution
		ItemSet *pattern = ItemSet::CreateSingleton(Uint32ItemSetType, seedNodes[r], numNodes);
		uint32 size = 1;
		connectedNodes->CleanTheSlate();
		connectedNodes->Unite(mDatabase->GetRow(seedNodes[r]));

		// With uniform background, all on level2 have the same score, which implies that the 
				// second node is always the highest-degree node connected to the initial node

		// Keep improving until max size is reached, or a local maximum -> will break in that case
		while(size < mMaxSize) {
			ItemSet *bestCandidate = NULL;
			double bestLambda = pattern->GetLambda();
			uint32 bestChange = UINT32_MAX_VALUE;
			bool bestIsExtension = true;

			// Foreach candidate node: add it
			for(uint32 c=0; c<numNodes; c++) {
				// Check if it not already there
				if(pattern->IsItemInSet(c))
					continue;

				// Is it connected to the current pattern?
				if(!connectedNodes->IsItemInSet(c))
					continue;

				// Construct candidate pattern
				ItemSet *candidate = ExtendPattern(pattern, c, size+1);
				double lambda = candidate->GetLambda();

				if(lambda > bestLambda) {		// Ties? --> First is highest-degree node and that is therefore currently the preferred one
					delete bestCandidate;					// Note: not necessarily true when removing nodes when iterative mining by removing nodes completely..
					bestCandidate = candidate;
					bestLambda = candidate->GetLambda();
					bestChange = c;
				} else
					delete candidate;

				++mCandidateCount;
			}

			// Check whether we found something
			if(bestCandidate==NULL) {		// No improved extension; local optimum found, try removing a node and otherwise break
				if(size==1)
					break;	// Cannot remove everything

				// Get ready for node removal
				bestIsExtension = false;
				uint32 *nodes = pattern->GetSet();

				// Foreach node in pattern
				for(uint32 c=0; c<size; c++) {
					// Construct candidate pattern
					ItemSet *candidate = ShrinkPattern(pattern, nodes[c], size-1);
					double lambda = candidate->GetLambda();

					if(lambda >= bestLambda) {
						delete bestCandidate;
						bestCandidate = candidate;
						bestLambda = candidate->GetLambda();
						bestChange = nodes[c];
					} else
						delete candidate;

					++mCandidateCount;
				}

				if(bestCandidate==NULL)	// Still nothing? -> Stop, this is the best we can do
					break; 
			} 

			// No break? -> We found an improved candidate, so use it!
			delete pattern;
			pattern = bestCandidate;
			size = pattern->GetLength();

			// Maintain the set of nodes connected to the current parent
			if(bestIsExtension)
				connectedNodes->Unite(mDatabase->GetRow(bestChange));
			else {
				// Rebuild connectedNodes from scratch (not very efficient, but shouldn't happen too often)
				connectedNodes->CleanTheSlate();
				uint32 *nodes = pattern->GetSet();
				for(uint32 i=0; i<size; i++)
					connectedNodes->Unite(mDatabase->GetRow(nodes[i]));
			}
		}

		// Add final result to collection (ie local optimum found)
		mPatternCollection->Add(pattern);
		delete pattern; // a clone is stored when needed
	}

	if(OutOfTime())
		LOG_MSG("** Out of time! Stopping right here, right now. **");

	// Cleanup
	delete connectedNodes;
	delete[] seedNodes;

	// Return #candidates visited
	return mCandidateCount;
}
