#include "../global.h"

#if defined (_WINDOWS)
	#include <direct.h>
#endif
#include <time.h>

// qtils
#include <Config.h>
#include <FileUtils.h>
#include <RandomUtils.h>
#include <StringUtils.h>
#include <TimeUtils.h>
#include <logger/Log.h>

// bass
#include <Bass.h>
#include <db/Database.h>
#include <itemstructs/PatternFile.h>
#include <collections/PatternCollection.h>

#include "../FicMain.h"
#include "../FicConf.h"
#include "../search/Search.h"
#include "../interest/Interestingness.h"

#include "MainTH.h"

MainTH::MainTH(Config *conf) : TaskHandler(conf){
}

MainTH::~MainTH() {
	// not my Config*
}

void MainTH::HandleTask() {
	string command = mConfig->Read<string>("command");
	
	if (command.compare("densesubgraph") == 0)				DoIt();
	else	if (command.compare("erscale") == 0)			ERScalability();

	else	throw string("MainTH :: Unable to handle task `" + command + "`");
}

string MainTH::BuildWorkingDir() {
	if (mConfig->Read<string>("command").compare("densesubgraph") == 0)
		return mConfig->Read<string>("command") + "/" + mConfig->Read<string>("searchType") + "/" + mConfig->Read<string>("dbName") + "-" + Logger::GetStartTimeString() + "/";
	else
		return mConfig->Read<string>("command") + "/";
}

/* ------------------------- Commands -------------------------- */

void MainTH::DoIt() {
	// Seed
	uint32 seed = mConfig->Read<uint32>("seed", 0);
	if(seed == 0)
		RandomUtils::Init();
	else
		RandomUtils::Init(seed);

	// Read DB
	Database *db = Database::RetrieveDatabase(mConfig->Read<string>("dbName"), Uint32ItemSetType, DbFileTypeUno);
	string dbName = db->GetName();

	// Init pattern collection and results file
	string patternsFilename = Bass::GetWorkingDir() + dbName + "-" + Logger::GetStartTimeString() + ".txt";
	PatternFile *pFile = new PatternFile(patternsFilename, db);
	PatternCollection *pCollection = PatternCollection::Create(mConfig, db, pFile);

	// Perform search
	printf_s("\n\n");
	Search *search = Search::Create(mConfig, db, pCollection);
	search->LogParameters();
	uint64 numCandidates = 0;
	Timer *timer = new Timer();
	for(uint32 i=0; i<pCollection->GetNumIterations(); i++) {
		printf_s(">> Searching pattern %u\r", i+1);
		numCandidates += search->DoSearch();
		pCollection->NextIteration();
		if(i+1<pCollection->GetNumIterations()) // at least one more iteration to come
			search->GetInterestingness()->AddToBackground(pCollection->GetPatterns()[pCollection->GetNumPatterns()-1]);
	}
	uint64 elapsedTime = timer->GetElapsedTime();
	pCollection->Finalize();
	printf_s("\n");

	// Log stats
	char temp[40]; sprintf_s(temp, 40, "%I64u", numCandidates);
	LOG_MSG("\nNumber of candidates: " + string(temp));
	sprintf_s(temp, 40, "%I64u", pCollection->GetNumPatterns());
	LOG_MSG("Number of patterns:   " + string(temp));

	// Compute some more statistics and save to file
	if(pCollection->IsLoaded()) {
		// Individual pattern statistics
		string statsFilename = Bass::GetWorkingDir() + "patterns-" + Logger::GetStartTimeString() + ".csv";
		pCollection->WritePatternDetails(statsFilename);

		// Global (pattern set) statistics
		FILE *fStats;
		statsFilename = Bass::GetWorkingDir() + "stats.txt";
		if(fopen_s(&fStats, statsFilename.c_str(), "w"))
			THROW("Could not write pattern set stats.");
		fprintf_s(fStats, "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%u\t", 
			Logger::GetStartTimeString().c_str(), mConfig->Read<string>("dbName").c_str(), mConfig->Read<string>("interestingness").c_str(), mConfig->Read<string>("numPatterns").c_str(),
			mConfig->Read<string>("maxSize", "").c_str(), mConfig->Read<string>("descriptionQ").c_str(), mConfig->Read<string>("searchType").c_str(), mConfig->Read<string>("bestMaxQueue", "").c_str(), 
			mConfig->Read<string>("seedMode", "uniform").c_str(), search->GetNumRunsPerIteration()
		);
		if(pCollection->GetNumPatterns() > 0) {
			pCollection->ComputeGlobalStats();
			fprintf_s(fStats, "%I64u\t%I64u\t%I64u\t%f\t%f\t",
				pCollection->GetNumPatterns(), numCandidates, elapsedTime, (double)pCollection->GetCoverage()/db->GetNumRows(), pCollection->GetAvgJaccard() 
			);
			ItemSet *top1 = pCollection->GetPattern(0);
			uint32 numNodes = top1->GetLength();
			SubgraphStats stats = top1->ComputeSubgraphStats(db);
			fprintf_s(fStats, "%.03f\t%.03f\t%.03f\t%u\t%.03f\t%.02f\n", top1->GetLambda(),
				stats.relSize, stats.edgeDensity, stats.diameter, stats.triangleDensity, stats.avgDegree);
		} else {
			fprintf_s(fStats, "%I64u\t%I64u\t%I64u\n",
				pCollection->GetNumPatterns(), numCandidates, elapsedTime
			);
		}
		fclose(fStats);
	}

	// Cleanup
	delete timer;
	delete search;
	delete pCollection;
	delete db;
}

void MainTH::ERScalability() {
	// Seed
	uint32 seed = mConfig->Read<uint32>("seed", 0);
	if (seed == 0)
		seed = (uint32)time(NULL);
	RandomUtils::Init(seed);

	// Read DB
	uint32 numNodes = mConfig->Read<uint32>("numNodes");
	double density = mConfig->Read<double>("density");
	Database *db = Database::GenerateErdosRenyi(numNodes, density);

	// Init pattern collection and results file
	//sprintf_s(temp, 100, "%u-%.4f-%u", numNodes, density, seed);
	//string patternsFilename = Bass::GetWorkingDir() + "er-" + temp + ".txt";
	//PatternFile *pFile = new PatternFile(patternsFilename, db);
	PatternCollection *pCollection = PatternCollection::Create(mConfig, db, NULL);

	// Perform search
	printf_s("\n\n");
	Search *search = Search::Create(mConfig, db, pCollection);
	search->LogParameters();
	uint64 numCandidates = 0;
	Timer *timer = new Timer();
	for (uint32 i = 0; i<pCollection->GetNumIterations(); i++) {
		printf_s(">> Searching pattern %u\r", i + 1);
		numCandidates += search->DoSearch();
		pCollection->NextIteration();
		if (i + 1<pCollection->GetNumIterations()) // at least one more iteration to come
			search->GetInterestingness()->AddToBackground(pCollection->GetPatterns()[pCollection->GetNumPatterns() - 1]);
	}
	uint64 elapsedTime = timer->GetElapsedTime();
	pCollection->Finalize();
	printf_s("\n");

	// Log stats
	char temp[40]; sprintf_s(temp, 40, "%I64u", numCandidates);
	LOG_MSG("\nNumber of candidates: " + string(temp));
	sprintf_s(temp, 40, "%I64u", pCollection->GetNumPatterns());
	LOG_MSG("Number of patterns:   " + string(temp));

		// Save performance stats
		FILE *fStats;
		sprintf_s(temp, 40, "%.04f", density);
		string savename = Bass::GetWorkingDir() + "er-perf-"+temp+".txt";
		if (fopen_s(&fStats, savename.c_str(), "a"))
			THROW("Could not write results.");
		fprintf_s(fStats, "%u\t%u\t%.4f\t%I64u\t%I64u\t%I64u\t%I64u\n", 
			seed, numNodes, density, db->GetNumItems(),
			pCollection->GetNumPatterns(), numCandidates, elapsedTime
		);
		fclose(fStats);

	// Cleanup
	delete timer;
	delete search;
	delete pCollection;
	delete db;
}
